import re

import databricks.sdk.useragent as ua
from databricks.labs.blueprint.logger import install_logger
from databricks.labs.ucx.__about__ import __version__

install_logger()

ua.semver_pattern = re.compile(
    r"^"
    r"(?P<major>0|[1-9]\d*)\.(?P<minor>x|0|[1-9]\d*)(\.(?P<patch>x|0|[1-9x]\d*))?"
    r"(?:-(?P<pre_release>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)"
    r"(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?"
    r"(?:\+(?P<build>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
)

# Add ucx/<version> for projects depending on ucx as a library
ua.with_extra("ucx", __version__)

# Add ucx/<version> for re-packaging of ucx, where product name is omitted
ua.with_product("ucx", __version__)



import dataclasses
from abc import ABC, abstractmethod
from dataclasses import dataclass


@dataclass
class TableIdentifier:
    catalog: str
    schema: str
    table: str

    @property
    def catalog_escaped(self):
        return f"`{self.catalog.replace('`','``')}`"

    @property
    def schema_escaped(self):
        return f"`{self.schema.replace('`','``')}`"

    @property
    def table_escaped(self):
        return f"`{self.table.replace('`','``')}`"

    @property
    def fqn_escaped(self):
        return f"{self.catalog_escaped}.{self.schema_escaped}.{self.table_escaped}"


@dataclass(frozen=True)
class ColumnMetadata:
    name: str
    data_type: str


@dataclass
class TableMetadata:
    identifier: TableIdentifier
    columns: list[ColumnMetadata]

    def get_column_metadata(self, column_name: str) -> ColumnMetadata | None:
        for column in self.columns:
            if column.name == column_name:
                return column
        return None


@dataclass
class DataProfilingResult:
    row_count: int
    table_metadata: TableMetadata


@dataclass
class SchemaComparisonEntry:
    source_column: str | None
    source_datatype: str | None
    target_column: str | None
    target_datatype: str | None
    is_matching: bool
    notes: str | None


@dataclass
class SchemaComparisonResult:
    is_matching: bool
    data: list[SchemaComparisonEntry]

    def as_dict(self):
        return dataclasses.asdict(self)


@dataclass
class DataComparisonResult:
    source_row_count: int
    target_row_count: int
    target_missing_count: int = 0
    source_missing_count: int = 0

    def as_dict(self):
        return dataclasses.asdict(self)


class TableMetadataRetriever(ABC):
    @abstractmethod
    def get_metadata(self, entity: TableIdentifier) -> TableMetadata:
        """
        Get metadata for a given table
        """


class DataProfiler(ABC):
    @abstractmethod
    def profile_data(self, entity: TableIdentifier) -> DataProfilingResult:
        """
        Profile data for a given table
        """


class SchemaComparator(ABC):
    @abstractmethod
    def compare_schema(self, source: TableIdentifier, target: TableIdentifier) -> SchemaComparisonResult:
        """
        Compare schema for two tables
        """


class DataComparator(ABC):
    @abstractmethod
    def compare_data(
        self,
        source: TableIdentifier,
        target: TableIdentifier,
        compare_rows: bool,
    ) -> DataComparisonResult:
        """
        Compare data for two tables
        """



import urllib.parse
import time
from collections.abc import Iterable
from dataclasses import dataclass, field
from datetime import timedelta
from typing import Any

from databricks.sdk.core import (
    ApiClient,
    AzureCliTokenSource,
    Config,
    CredentialsProvider,
    CredentialsStrategy,
    credentials_strategy,
)
from databricks.sdk.errors import NotFound, PermissionDenied, ResourceConflict
from databricks.sdk.retries import retried

from databricks.labs.ucx.assessment.crawlers import logger

# https://learn.microsoft.com/en-us/azure/role-based-access-control/built-in-roles
_ROLES = {
    "STORAGE_BLOB_DATA_READER": "2a2b9908-6ea1-4ae2-8e65-a410df84e7d1",
    "STORAGE_BLOB_DATA_CONTRIBUTOR": "ba92f5b4-2d11-453d-a403-e96b0029c9fe",
}


@dataclass
class AzureSubscription:
    name: str
    subscription_id: str
    tenant_id: str


class AzureResource:
    def __init__(self, resource_id: str):
        self._pairs = {}
        self._resource_id = resource_id
        split = resource_id.lstrip("/").split("/")
        if len(split) % 2 != 0:
            msg = f"not a list of pairs: {resource_id}"
            raise ValueError(msg)
        i = 0
        while i < len(split):
            k = split[i]
            value = split[i + 1]
            i += 2
            self._pairs[k] = value

    @property
    def subscription_id(self):
        return self._pairs.get("subscriptions")

    @property
    def resource_group(self):
        return self._pairs.get("resourceGroups")

    @property
    def storage_account(self):
        return self._pairs.get("storageAccounts")

    @property
    def container(self):
        return self._pairs.get("containers")

    @property
    def access_connector(self):
        return self._pairs.get("accessConnectors")

    def __eq__(self, other):
        if not isinstance(other, AzureResource):
            return NotImplemented
        return self._resource_id == other._resource_id

    def __repr__(self):
        properties = ["subscription_id", "resource_group", "storage_account", "container", "access_connector"]
        pairs = [f"{_}={getattr(self, _)}" for _ in properties]
        return f'AzureResource<{", ".join(pairs)}>'

    def __str__(self):
        return self._resource_id


class RawResource:
    def __init__(self, raw_resource: dict[str, Any]):
        if "id" not in raw_resource:
            raise KeyError("Raw resource must contain an 'id' field")
        self._id = AzureResource(raw_resource["id"])
        self._raw_resource = raw_resource

    @property
    def id(self) -> AzureResource:
        return self._id

    def get(self, key: str, default: Any) -> Any:
        return self._raw_resource.get(key, default)


@dataclass
class Principal:
    client_id: str
    display_name: str
    object_id: str
    # service principal will have type: "Application"
    # managed identity will have type: "ManagedIdentity"
    type: str
    # Need this directory_id/tenant_id when create UC storage credentials using service principal
    # it will be None if type is managed identity
    directory_id: str | None = None


@dataclass
class StorageAccount:
    id: AzureResource
    name: str
    location: str
    default_network_action: str  # "Unknown", "Deny" or "Allow"

    @classmethod
    def from_raw_resource(cls, raw: RawResource) -> "StorageAccount":
        if raw.id is None:
            raise KeyError(f"Missing id: {raw}")

        name = raw.get("name", "")
        if name == "":
            raise KeyError(f"Missing name: {raw}")

        location = raw.get("location", "")
        if location == "":
            raise KeyError(f"Missing location: {raw}")

        default_network_action = raw.get("properties", {}).get("networkAcls", {}).get("defaultAction", "Unknown")

        storage_account = cls(id=raw.id, name=name, location=location, default_network_action=default_network_action)
        return storage_account


@dataclass
class PrincipalSecret:
    client: Principal
    secret: str


@dataclass
class AzureRoleAssignment:
    id: str
    resource: AzureResource
    scope: AzureResource
    principal: Principal
    role_name: str
    role_type: str
    role_permissions: list[str]


@dataclass
class AzureRoleDetails:
    role_name: str | None
    role_type: str
    role_permissions: list[str]


@dataclass
class AccessConnector:
    id: AzureResource
    name: str
    location: str
    provisioning_state: str  # https://learn.microsoft.com/en-us/rest/api/databricks/access-connectors/get?view=rest-databricks-2023-05-01&tabs=HTTP#provisioningstate
    identity_type: str  # SystemAssigned or UserAssigned
    principal_id: str
    managed_identity_type: str | None = None  # str when identity_type is UserAssigned
    client_id: str | None = None  # str when identity_type is UserAssigned
    tenant_id: str | None = None  # str when identity_type is SystemAssigned
    tags: dict[str, str] = field(default_factory=dict)

    @classmethod
    def from_raw_resource(cls, raw: RawResource) -> "AccessConnector":
        if raw.id is None:
            raise KeyError(f"Missing id: {raw}")

        name = raw.get("name", "")
        if name == "":
            raise KeyError(f"Missing name: {raw}")

        location = raw.get("location", "")
        if location == "":
            raise KeyError(f"Missing location: {raw}")

        provisioning_state = raw.get("properties", {}).get("provisioningState", "")
        if provisioning_state == "":
            raise KeyError(f"Missing provisioning state: {raw}")

        identity = raw.get("identity", {})
        identity_type = identity.get("type")
        if identity_type == "UserAssigned":
            if len(identity.keys()) > 1:
                raise KeyError(f"Multiple user assigned identities: {identity.keys()}")
            if len(identity.keys()) == 0:
                raise KeyError(f"No user assigned identity: {identity.keys()}")
            managed_identity_id = list(identity.keys())[0]
            principal_id = identity[managed_identity_id]["principalId"]
            client_id = identity[managed_identity_id]["clientId"]
            tenant_id = None
        elif identity_type == "SystemAssigned":
            principal_id = identity["principalId"]
            managed_identity_id = client_id = None
            tenant_id = identity["tenantId"]
        else:
            raise KeyError(f"Unsupported identity type: {identity_type}")

        access_connector = cls(
            id=raw.id,
            name=name,
            location=location,
            provisioning_state=provisioning_state,
            identity_type=identity_type,
            principal_id=principal_id,
            managed_identity_type=managed_identity_id,
            client_id=client_id,
            tenant_id=tenant_id,
            tags=raw.get("tags", {}),
        )
        return access_connector


class AzureAPIClient:
    def __init__(self, host_endpoint: str, service_endpoint: str):
        self.api_client = ApiClient(
            Config(
                host=host_endpoint,
                credentials_strategy=self._strategy_for(service_endpoint),
            )
        )
        self._token_source = AzureCliTokenSource(host_endpoint)

    @staticmethod
    def _strategy_for(endpoint: str) -> CredentialsStrategy:
        @credentials_strategy("azure-cli", ["host"])
        def _credentials(_: Config) -> CredentialsProvider:
            token_source = AzureCliTokenSource(endpoint)

            def inner() -> dict[str, str]:
                token = token_source.token()
                return {"Authorization": f"{token.token_type} {token.access_token}"}

            return inner

        return _credentials

    def get(self, path: str, api_version: str | None = None, query: dict[str, str] | None = None):
        headers = {"Accept": "application/json"}
        _query: dict[str, str] = query or {}
        if api_version is not None:
            _query["api-version"] = api_version
        return self.api_client.do("GET", path, query=_query, headers=headers)

    def put(self, path: str, api_version: str | None = None, body: dict | None = None):
        headers = {"Content-Type": "application/json"}
        query: dict[str, str] = {}
        if api_version is not None:
            query["api-version"] = api_version
        if body is not None:
            return self.api_client.do("PUT", path, query=query, headers=headers, body=body)
        return None

    def post(self, path: str, body: dict | None = None):
        headers = {"Content-Type": "application/json"}
        query: dict[str, str] = {}
        if body is not None:
            return self.api_client.do("POST", path, query=query, headers=headers, body=body)
        return self.api_client.do("POST", path, query=query, headers=headers)

    def delete(self, path: str, api_version: str | None = None):
        # this method is added only to be used in int test to delete the application once tests pass
        headers = {"Content-Type": "application/json"}
        query: dict[str, str] = {}
        if api_version is not None:
            query["api-version"] = api_version
        return self.api_client.do("DELETE", path, query=query, headers=headers)

    def token(self):
        return self._token_source.token()


class AzureResources:
    def __init__(self, azure_mgmt: AzureAPIClient, azure_graph: AzureAPIClient, include_subscriptions=None):
        if not include_subscriptions:
            include_subscriptions = []
        self._mgmt = azure_mgmt
        self._graph = azure_graph
        self._include_subscriptions = include_subscriptions
        self._role_definitions = {}  # type: dict[str, AzureRoleDetails]
        self._principals: dict[str, Principal | None] = {}

    def _get_subscriptions(self) -> Iterable[AzureSubscription]:
        for subscription in self._mgmt.get("/subscriptions", "2022-12-01").get("value", []):
            yield AzureSubscription(
                name=subscription["displayName"],
                subscription_id=subscription["subscriptionId"],
                tenant_id=subscription["tenantId"],
            )

    def create_service_principal(self, display_name: str) -> PrincipalSecret:
        try:
            application_info: dict[str, str] = self._graph.post("/v1.0/applications", {"displayName": display_name})
            app_id = application_info.get("appId")
            assert app_id is not None
            service_principal_info: dict[str, str] = self._graph.post("/v1.0/servicePrincipals", {"appId": app_id})
            object_id = service_principal_info.get("id")
            assert object_id is not None
            secret_info: dict[str, str] = self._graph.post(f"/v1.0/servicePrincipals/{object_id}/addPassword")

        except PermissionDenied:
            msg = (
                "Permission denied. Please run this cmd under the identity of a user who has"
                " create service principal permission."
            )
            logger.error(msg)
            raise PermissionDenied(msg) from None
        secret = secret_info.get("secretText")
        client_id = service_principal_info.get("appId")
        principal_type = service_principal_info.get("servicePrincipalType")
        directory_id = service_principal_info.get("appOwnerOrganizationId")
        assert client_id is not None
        assert object_id is not None
        assert principal_type is not None
        assert directory_id is not None
        assert secret is not None
        principal_secret = PrincipalSecret(
            Principal(client_id, display_name, object_id, principal_type, directory_id), secret
        )
        logger.info(
            f"Created service principal ({principal_secret.client.client_id}) with access to used storage accounts: "
            + principal_secret.client.display_name
        )
        return principal_secret

    def delete_service_principal(self, principal_id: str, *, safe: bool = False):
        """Delete the service principal.

        Parameters
        ----------
        principal_id : str
            The principal id to delete.
        safe : bool, optional (default: True)
            If True, will not raise an error if the service principal does not exists.

        Raises
        ------
        NotFound :
            If the principal is not found.
        PermissionDenied :
            If missing permission to delete the service principal
        """
        try:
            self._graph.delete(f"/v1.0/applications(appId='{principal_id}')")
        except PermissionDenied:
            msg = f"User doesnt have permission to delete application {principal_id}"
            logger.error(msg)
            raise PermissionDenied(msg) from None
        except NotFound:
            if safe:
                return
            raise

    def _log_permission_denied_error_for_storage_permission(self, path: str) -> None:
        logger.error(
            "Permission denied. Please run this cmd under the identity of a user who has "
            f"create service principal permission: {path}"
        )

    def get_storage_permission(
        self,
        storage_account: StorageAccount,
        role_guid: str,
        *,
        timeout: timedelta = timedelta(seconds=1),
    ) -> AzureRoleAssignment | None:
        """Get a storage permission.

        Parameters
        ----------
        storage_account : StorageAccount
            The storage account to get the permission for.
        role_guid : str
            The role guid to get the permission for.
        timeout : timedelta, optional (default: timedelta(seconds=1))
            The timeout to wait for the permission to be found.

        Raises
        ------
        PermissionDenied :
            If user is missing permission to get the storage permission.
        """
        retry = retried(on=[NotFound], timeout=timeout)
        path = f"{storage_account.id}/providers/Microsoft.Authorization/roleAssignments/{role_guid}"
        try:
            response = retry(self._mgmt.get)(path, "2022-04-01")
            assignment = self._role_assignment(response, str(storage_account.id))
            return assignment
        except TimeoutError:  # TimeoutError is raised by retried
            logger.warning(f"Storage permission not found: {path}")  # not found because retry on NotFound
            return None
        except PermissionDenied:
            self._log_permission_denied_error_for_storage_permission(path)
            raise

    def _get_storage_permissions(
        self,
        principal_id: str,
        storage_account: StorageAccount,
    ) -> Iterable[AzureRoleAssignment]:
        """Get storage permissions for a principal.

        Parameters
        ----------
        principal_id : str
            The principal id to get the storage permissions for.
        storage_account : StorageAccount
            The storage account to get the permission for.

        Yields
        ------
        AzureRoleAssignment :
            The role assignment

        Raises
        ------
        PermissionDenied :
            If user is missing permission to get the storage permission.
        """
        path = (
            f"{storage_account.id}/providers/Microsoft.Authorization/roleAssignments"
            f"?$filter=principalId%20eq%20'{principal_id}'"
        )
        try:
            response = self._mgmt.get(path, "2022-04-01")
        except PermissionDenied:
            self._log_permission_denied_error_for_storage_permission(path)
            raise

        for role_assignment in response.get("value", []):
            assignment = self._role_assignment(role_assignment, str(storage_account.id))
            if assignment:
                yield assignment

    def apply_storage_permission(
        self, principal_id: str, storage_account: StorageAccount, role_name: str, role_guid: str
    ):
        role_id = _ROLES[role_name]
        path = f"{storage_account.id}/providers/Microsoft.Authorization/roleAssignments/{role_guid}"
        try:
            role_definition_id = f"/subscriptions/{storage_account.id.subscription_id}/providers/Microsoft.Authorization/roleDefinitions/{role_id}"
            body = {
                "properties": {
                    "roleDefinitionId": role_definition_id,
                    "principalId": principal_id,
                    "principalType": "ServicePrincipal",
                }
            }
            self._mgmt.put(path, "2022-04-01", body)
        except ResourceConflict:
            logger.warning(
                f"Role assignment already exists for role {role_guid} on storage {storage_account.name}"
                f" for spn {principal_id}."
            )
        except PermissionDenied:
            self._log_permission_denied_error_for_storage_permission(path)
            raise

    def _delete_storage_permission(
        self, principal_id: str, storage_account: StorageAccount, *, safe: bool = False
    ) -> None:
        """See meth:delete_storage_permission"""
        try:
            storage_permissions = list(self._get_storage_permissions(principal_id, storage_account))
        except NotFound:
            if safe:
                return
            raise
        permission_denied_ids = []
        for permission in storage_permissions:
            try:
                self._mgmt.delete(permission.id, "2022-04-01")
            except PermissionDenied:
                self._log_permission_denied_error_for_storage_permission(permission.id)
                permission_denied_ids.append(permission.id)
            except NotFound:
                continue  # Somehow deleted right in-between getting and deleting
        if permission_denied_ids:
            raise PermissionDenied(
                f"Permission denied for deleting role assignments: {', '.join(permission_denied_ids)}"
            )

    def delete_storage_permission(
        self, principal_id: str, *storage_accounts: StorageAccount, safe: bool = False
    ) -> None:
        """Delete storage permission(s) for a principal

        Parameters
        ----------
        principal_id : str
            The principal id to delete the role assignment(s) for.
        storage_accounts : StorageAccount
            The storage account(s) to delete permission for.
        safe : bool, optional (default: False)
            If True, will not raise an exception if no role assignment are found.

        Raises
        ------
        PermissionDenied :
            If user is missing permission to get the storage permission.
        """
        for storage_account in storage_accounts:
            self._delete_storage_permission(principal_id, storage_account, safe=safe)

    def tenant_id(self):
        token = self._mgmt.token()
        return token.jwt_claims().get("tid")

    def subscriptions(self):
        tenant_id = self.tenant_id()
        for subscription in self._get_subscriptions():
            if subscription.tenant_id != tenant_id:
                continue
            if subscription.subscription_id not in self._include_subscriptions:
                continue
            yield subscription

    def storage_accounts(self) -> Iterable[StorageAccount]:
        for subscription in self.subscriptions():
            logger.info(f"Checking in subscription {subscription.name} for storage accounts")
            path = f"/subscriptions/{subscription.subscription_id}/providers/Microsoft.Storage/storageAccounts"
            for response in self._mgmt.get(path, "2023-01-01").get("value", []):
                try:
                    storage_account = StorageAccount.from_raw_resource(RawResource(response))
                except KeyError:
                    logger.warning(f"Failed parsing storage account: {response}")
                else:
                    yield storage_account

    def containers(self, storage: AzureResource):
        for raw in self._mgmt.get(f"{storage}/blobServices/default/containers", "2023-01-01").get("value", []):
            resource_id = raw.get("id")
            if not resource_id:
                continue
            yield AzureResource(resource_id)

    def _get_principal(self, principal_id: str) -> Principal | None:
        if principal_id in self._principals:
            return self._principals[principal_id]
        try:
            path = f"/v1.0/directoryObjects/{principal_id}"
            raw: dict[str, str] = self._graph.get(path)  # type: ignore[assignment]
        except NotFound:
            # don't load principals from external directories twice
            self._principals[principal_id] = None
            return self._principals[principal_id]

        client_id = raw.get("appId")
        display_name = raw.get("displayName")
        object_id = raw.get("id")
        principal_type = raw.get("servicePrincipalType")
        # Need this directory_id/tenant_id when create UC storage credentials using service principal
        directory_id = raw.get("appOwnerOrganizationId")
        assert client_id is not None
        assert display_name is not None
        assert object_id is not None
        assert principal_type is not None
        if principal_type == "Application":
            # service principal must have directory_id
            assert directory_id is not None
        self._principals[principal_id] = Principal(client_id, display_name, object_id, principal_type, directory_id)
        return self._principals[principal_id]

    def role_assignments(
        self, resource_id: str, *, principal_types: list[str] | None = None
    ) -> Iterable[AzureRoleAssignment]:
        """See https://learn.microsoft.com/en-us/rest/api/authorization/role-assignments/list-for-resource"""
        if not principal_types:
            principal_types = ["ServicePrincipal"]
        result = self._mgmt.get(f"{resource_id}/providers/Microsoft.Authorization/roleAssignments", "2022-04-01")
        for role_assignment in result.get("value", []):
            principal_type = role_assignment.get("properties", {}).get("principalType")
            if not principal_type or principal_type not in principal_types:
                continue
            assignment = self._role_assignment(role_assignment, resource_id)
            if not assignment:
                continue
            yield assignment

    def _role_assignment(self, role_assignment: dict, resource_id: str) -> AzureRoleAssignment | None:
        id_ = role_assignment.get("id")
        if not id_:
            return None
        assignment_properties = role_assignment.get("properties", {})
        principal_type = assignment_properties.get("principalType")
        if not principal_type:
            return None
        principal_id = assignment_properties.get("principalId")
        if not principal_id:
            return None
        role_definition_id = assignment_properties.get("roleDefinitionId")
        if not role_definition_id:
            return None
        scope = assignment_properties.get("scope")
        if not scope:
            return None
        role_details = self._role_name(role_definition_id)
        role_name = role_details.role_name
        if not role_name:
            return None
        principal = self._get_principal(principal_id)
        if not principal:
            return None
        if scope == "/":
            scope = resource_id
        return AzureRoleAssignment(
            id=id_,
            resource=AzureResource(resource_id),
            scope=AzureResource(scope),
            principal=principal,
            role_name=role_name,
            role_type=role_details.role_type,
            role_permissions=role_details.role_permissions,
        )

    def _role_name(self, role_definition_id) -> AzureRoleDetails:
        if role_definition_id not in self._role_definitions:
            role_definition = self._mgmt.get(role_definition_id, "2022-04-01")
            definition_properties = role_definition.get("properties", {})
            role_name = definition_properties.get("roleName")
            if not role_name:
                return AzureRoleDetails(role_name=None, role_type='BuiltInRole', role_permissions=[])
            role_type = definition_properties.get("type", "BuiltInRole")
            role_permissions = []
            if role_type == 'CustomRole':
                role_permissions_list = definition_properties.get("permissions", [])
                for each_role_permissions in role_permissions_list:
                    role_permissions = each_role_permissions.get("actions", []) + each_role_permissions.get(
                        "dataActions", []
                    )
            self._role_definitions[role_definition_id] = AzureRoleDetails(
                role_name=role_name, role_type=role_type, role_permissions=role_permissions
            )
        return self._role_definitions[role_definition_id]

    def managed_identity_client_id(
        self, access_connector_id: str, user_assigned_identity_id: str | None = None
    ) -> str | None:
        # get te client_id/application_id of the managed identity used in the access connector
        try:
            identity = self._mgmt.get(access_connector_id, "2023-05-01").get("identity")
        except NotFound:
            logger.warning(f"Access connector {access_connector_id} no longer exists")
            return None
        if not identity:
            return None

        if identity.get("type") == "UserAssigned":
            if not user_assigned_identity_id:
                return None
            identities = identity.get("userAssignedIdentities")
            if user_assigned_identity_id in identities:
                return identities.get(user_assigned_identity_id).get("clientId")
            # sometimes we see "resourceGroups" instead of "resourcegroups" in the response from Azure RM API
            # but "resourcegroups" is in response from storage credential's managed_identity_id
            alternative_identity_id = user_assigned_identity_id.replace("resourcegroups", "resourceGroups")
            if alternative_identity_id in identities:
                return identities.get(alternative_identity_id).get("clientId")
            return None
        if identity.get("type") == "SystemAssigned":
            # SystemAssigned managed identity does not have client_id in get access connector response
            # need to call graph api directoryObjects to fetch the client_id
            principal = self._get_principal(identity.get("principalId"))
            if not principal:
                return None
            return principal.client_id
        return None

    def get_access_connector(self, subscription_id: str, resource_group_name: str, name: str) -> AccessConnector | None:
        """Get an access connector.

        Docs:
            https://learn.microsoft.com/en-us/rest/api/databricks/access-connectors/get?view=rest-databricks-2023-05-01&tabs=HTTP
        """
        url = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Databricks/accessConnectors/{name}"
        response = self._mgmt.get(url, api_version="2023-05-01")
        raw = RawResource(response)
        try:
            access_connector = AccessConnector.from_raw_resource(raw)
        except KeyError:
            logger.warning(f"Tried getting non-existing access connector: {url}")
            access_connector = None
        return access_connector

    def list_resources(self, subscription: AzureSubscription, resource_type: str) -> Iterable[RawResource]:
        """List all resources of a type within subscription"""
        query = {"api-version": "2020-06-01", "$filter": f"resourceType eq '{resource_type}'"}
        while True:
            res = self._mgmt.get(f"/subscriptions/{subscription.subscription_id}/resources", query=query)
            for resource in res["value"]:
                try:
                    yield RawResource(resource)
                except KeyError:
                    logger.warning(f"Could not parse resource: {resource}")

            next_link = res.get("nextLink", None)
            if not next_link:
                break
            parsed_link = urllib.parse.urlparse(next_link)
            query = dict(urllib.parse.parse_qsl(parsed_link.query))

    def access_connectors(self) -> Iterable[AccessConnector]:
        """List all access connector within subscription

        Docs:
            https://learn.microsoft.com/en-us/rest/api/databricks/access-connectors/list-by-subscription?view=rest-databricks-2023-05-01&tabs=HTTP
        """
        for subscription in self.subscriptions():
            for raw in self.list_resources(subscription, "Microsoft.Databricks/accessConnectors"):
                try:
                    yield AccessConnector.from_raw_resource(raw)
                except KeyError:
                    logger.warning(f"Could not parse access connector: {raw}")

    def create_or_update_access_connector(
        self,
        subscription_id: str,
        resource_group_name: str,
        name: str,
        location: str,
        tags: dict[str, str] | None,
        *,
        wait_for_provisioning: bool = False,
        wait_for_provisioning_timeout_in_seconds: int = 300,
    ) -> AccessConnector:
        """Create access connector.

        Docs:
            https://learn.microsoft.com/en-us/rest/api/databricks/access-connectors/create-or-update?view=rest-databricks-2023-05-01&tabs=HTTP
        """
        url = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Databricks/accessConnectors/{name}"
        body = {
            "location": location,
            "identity": {"type": "SystemAssigned"},
        }
        if tags is not None:
            body["tags"] = tags
        self._mgmt.put(url, api_version="2023-05-01", body=body)

        access_connector = self.get_access_connector(subscription_id, resource_group_name, name)

        start_time = time.time()
        if access_connector is None or (wait_for_provisioning and access_connector.provisioning_state != "Succeeded"):
            if time.time() - start_time > wait_for_provisioning_timeout_in_seconds:
                raise TimeoutError(f"Timeout waiting for creating or updating access connector: {url}")
            time.sleep(5)

            access_connector = self.get_access_connector(subscription_id, resource_group_name, name)
            assert access_connector is not None

        return access_connector

    def delete_access_connector(self, url: str) -> None:
        """Delete an access connector.

        Docs:
            https://learn.microsoft.com/en-us/rest/api/databricks/access-connectors/delete?view=rest-databricks-2023-05-01&tabs=HTTP
        """
        self._mgmt.delete(url, api_version="2023-05-01")



from databricks.labs.ucx.contexts.workflow_task import RuntimeContext
from databricks.labs.ucx.framework.tasks import Workflow, job_task


class ExperimentalWorkflowLinter(Workflow):
    def __init__(self):
        super().__init__('experimental-workflow-linter')

    @job_task(job_cluster="table_migration")
    def lint_all_workflows(self, ctx: RuntimeContext):
        """[EXPERIMENTAL] Analyses all jobs for source code compatibility problems. This is an experimental feature,
        that is not yet fully supported."""
        ctx.workflow_linter.refresh_report(ctx.sql_backend, ctx.inventory_database)

    @job_task(job_cluster="table_migration")
    def lint_all_queries(self, ctx: RuntimeContext):
        """[EXPERIMENTAL] Analyses all jobs for source code compatibility problems. This is an experimental feature,
        that is not yet fully supported."""
        ctx.query_linter.refresh_report(ctx.sql_backend, ctx.inventory_database)



from __future__ import annotations

import dataclasses
import logging
import sys
from collections.abc import Sequence, Iterable
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any

from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk.errors import DatabricksError

from databricks.labs.ucx.framework.utils import escape_sql_identifier

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


logger = logging.getLogger(__name__)


@dataclass
class LineageAtom:

    object_type: str
    object_id: str
    other: dict[str, str] | None = None


@dataclass
class DirectFsAccess:
    """A record describing a Direct File System Access"""

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        source_lineage = data.get("source_lineage", None)
        if isinstance(source_lineage, list) and len(source_lineage) > 0 and isinstance(source_lineage[0], dict):
            lineage_atoms = [LineageAtom(**lineage) for lineage in source_lineage]
            data["source_lineage"] = lineage_atoms
        return cls(**data)

    UNKNOWN = "unknown"

    path: str
    is_read: bool
    is_write: bool
    source_id: str = UNKNOWN
    source_timestamp: datetime = datetime.fromtimestamp(0)
    source_lineage: list[LineageAtom] = field(default_factory=list)
    assessment_start_timestamp: datetime = datetime.fromtimestamp(0)
    assessment_end_timestamp: datetime = datetime.fromtimestamp(0)

    def replace_source(
        self,
        source_id: str | None = None,
        source_lineage: list[LineageAtom] | None = None,
        source_timestamp: datetime | None = None,
    ):
        return dataclasses.replace(
            self,
            source_id=source_id or self.source_id,
            source_timestamp=source_timestamp or self.source_timestamp,
            source_lineage=source_lineage or self.source_lineage,
        )

    def replace_assessment_infos(
        self, assessment_start: datetime | None = None, assessment_end: datetime | None = None
    ):
        return dataclasses.replace(
            self,
            assessment_start_timestamp=assessment_start or self.assessment_start_timestamp,
            assessment_end_timestamp=assessment_end or self.assessment_end_timestamp,
        )


class DirectFsAccessCrawler(CrawlerBase[DirectFsAccess]):

    @classmethod
    def for_paths(cls, backend: SqlBackend, schema) -> DirectFsAccessCrawler:
        return DirectFsAccessCrawler(backend, schema, "directfs_in_paths")

    @classmethod
    def for_queries(cls, backend: SqlBackend, schema) -> DirectFsAccessCrawler:
        return DirectFsAccessCrawler(backend, schema, "directfs_in_queries")

    def __init__(self, backend: SqlBackend, schema: str, table: str):
        """
        Initializes a DFSACrawler instance.

        Args:
            sql_backend (SqlBackend): The SQL Execution Backend abstraction (either REST API or Spark)
            schema: The schema name for the inventory persistence.
        """
        super().__init__(backend=backend, catalog="hive_metastore", schema=schema, table=table, klass=DirectFsAccess)

    def dump_all(self, dfsas: Sequence[DirectFsAccess]):
        """This crawler doesn't follow the pull model because the fetcher fetches data for 2 crawlers, not just one
        It's not **bad** because all records are pushed at once.
        Providing a multi-entity crawler is out-of-scope of this PR
        """
        try:
            # TODO until we historize data, we append all DFSAs
            self._update_snapshot(dfsas, mode="append")
        except DatabricksError as e:
            logger.error("Failed to store DFSAs", exc_info=e)

    def _try_fetch(self) -> Iterable[DirectFsAccess]:
        sql = f"SELECT * FROM {escape_sql_identifier(self.full_name)}"
        for row in self._backend.fetch(sql):
            yield self._klass.from_dict(row.as_dict())

    def _crawl(self) -> Iterable[DirectFsAccess]:
        return []
        # TODO raise NotImplementedError() once CrawlerBase supports empty snapshots



import json
import logging
from collections.abc import Iterable
from dataclasses import dataclass

from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient

from databricks.labs.ucx.assessment.clusters import CheckClusterMixin
from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier

logger = logging.getLogger(__name__)


@dataclass
class PipelineInfo:
    pipeline_id: str
    success: int
    failures: str
    pipeline_name: str | None = None
    creator_name: str | None = None


class PipelinesCrawler(CrawlerBase[PipelineInfo], CheckClusterMixin):
    def __init__(self, ws: WorkspaceClient, sbe: SqlBackend, schema):
        super().__init__(sbe, "hive_metastore", schema, "pipelines", PipelineInfo)
        self._ws = ws

    def _crawl(self) -> Iterable[PipelineInfo]:
        all_pipelines = list(self._ws.pipelines.list_pipelines())
        return list(self._assess_pipelines(all_pipelines))

    def _assess_pipelines(self, all_pipelines) -> Iterable[PipelineInfo]:
        for pipeline in all_pipelines:
            if not pipeline.creator_user_name:
                logger.warning(
                    f"Pipeline {pipeline.name} have Unknown creator, it means that the original creator "
                    f"has been deleted and should be re-created"
                )
            pipeline_info = PipelineInfo(
                pipeline_id=pipeline.pipeline_id,
                pipeline_name=pipeline.name,
                creator_name=pipeline.creator_user_name,
                success=1,
                failures="[]",
            )

            failures = []
            pipeline_response = self._ws.pipelines.get(pipeline.pipeline_id)
            assert pipeline_response.spec is not None
            pipeline_config = pipeline_response.spec.configuration
            if pipeline_config:
                failures.extend(self._check_spark_conf(pipeline_config, "pipeline"))
            clusters = pipeline_response.spec.clusters
            if clusters:
                self._pipeline_clusters(clusters, failures)
            pipeline_info.failures = json.dumps(failures)
            if len(failures) > 0:
                pipeline_info.success = 0
            yield pipeline_info

    def _pipeline_clusters(self, clusters, failures):
        for cluster in clusters:
            if cluster.spark_conf:
                failures.extend(self._check_spark_conf(cluster.spark_conf, "pipeline cluster"))
            # Checking if cluster config is present in cluster policies
            if cluster.policy_id:
                failures.extend(self._check_cluster_policy(cluster.policy_id, "pipeline cluster"))
            if cluster.init_scripts:
                failures.extend(self._check_cluster_init_script(cluster.init_scripts, "pipeline cluster"))

    def _try_fetch(self) -> Iterable[PipelineInfo]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield PipelineInfo(*row)



# pylint: disable=invalid-name

import logging

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.tui import Prompts
from databricks.sdk import WorkspaceClient

from databricks.labs.ucx.config import WorkspaceConfig
from databricks.labs.ucx.installer.policy import ClusterPolicyInstaller

logger = logging.getLogger(__name__)


def upgrade(installation: Installation, ws: WorkspaceClient):
    config = installation.load(WorkspaceConfig)
    policy_installer = ClusterPolicyInstaller(installation, ws, Prompts())
    config.policy_id, _, _, _ = policy_installer.create(config.inventory_database)
    installation.save(config)
    states = InstallState.from_installation(installation)
    assert config.policy_id is not None
    policy_installer.update_job_policy(states, config.policy_id)



import dataclasses
import functools
import logging
import shutil
import tempfile
from collections.abc import Generator, Iterable
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime, timezone
from importlib import metadata
from pathlib import Path
from urllib import parse

from databricks.labs.blueprint.parallel import ManyError, Threads
from databricks.labs.blueprint.paths import DBFSPath
from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound
from databricks.sdk.service import compute, jobs

from databricks.labs.ucx.assessment.crawlers import runtime_version_tuple
from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationIndex
from databricks.labs.ucx.mixins.cached_workspace_path import WorkspaceCache
from databricks.labs.ucx.source_code.base import (
    CurrentSessionState,
    LocatedAdvice,
    is_a_notebook,
    file_language,
    guess_encoding,
)
from databricks.labs.ucx.source_code.directfs_access import (
    LineageAtom,
    DirectFsAccessCrawler,
    DirectFsAccess,
)
from databricks.labs.ucx.source_code.graph import (
    Dependency,
    DependencyGraph,
    DependencyProblem,
    DependencyResolver,
    SourceContainer,
    WrappingLoader,
    DependencyGraphWalker,
)
from databricks.labs.ucx.source_code.linters.context import LinterContext
from databricks.labs.ucx.source_code.linters.directfs import DirectFsAccessPyLinter, DirectFsAccessSqlLinter
from databricks.labs.ucx.source_code.notebooks.cells import CellLanguage
from databricks.labs.ucx.source_code.python.python_ast import Tree
from databricks.labs.ucx.source_code.notebooks.sources import FileLinter, Notebook
from databricks.labs.ucx.source_code.path_lookup import PathLookup

logger = logging.getLogger(__name__)


@dataclass
class JobProblem:
    job_id: int
    job_name: str
    task_key: str
    path: str
    code: str
    message: str
    start_line: int
    start_col: int
    end_line: int
    end_col: int

    def as_message(self) -> str:
        message = f"{self.path}:{self.start_line} [{self.code}] {self.message}"
        return message


class WorkflowTask(Dependency):
    def __init__(self, ws: WorkspaceClient, task: jobs.Task, job: jobs.Job):
        loader = WrappingLoader(WorkflowTaskContainer(ws, task, job))
        super().__init__(loader, Path(f'/jobs/{task.task_key}'), inherits_context=False)
        self._task = task
        self._job = job

    def load(self, path_lookup: PathLookup) -> SourceContainer | None:
        return self._loader.load_dependency(path_lookup, self)

    def __repr__(self):
        return f'WorkflowTask<{self._task.task_key} of {self._job.settings.name}>'

    @property
    def lineage(self) -> list[LineageAtom]:
        job_name = (None if self._job.settings is None else self._job.settings.name) or "unknown job"
        job_lineage = LineageAtom("JOB", str(self._job.job_id), {"name": job_name})
        task_lineage = LineageAtom("TASK", self._task.task_key)
        return [job_lineage, task_lineage]


class WorkflowTaskContainer(SourceContainer):
    def __init__(self, ws: WorkspaceClient, task: jobs.Task, job: jobs.Job):
        self._task = task
        self._job = job
        self._ws = ws
        self._cache = WorkspaceCache(ws)
        self._named_parameters: dict[str, str] | None = {}
        self._parameters: list[str] | None = []
        self._spark_conf: dict[str, str] | None = {}
        self._spark_version: str | None = None
        self._data_security_mode = None
        self._is_serverless = False

    @property
    def named_parameters(self) -> dict[str, str]:
        return self._named_parameters or {}

    @property
    def spark_conf(self) -> dict[str, str]:
        return self._spark_conf or {}

    @property
    def runtime_version(self) -> tuple[int, int]:
        version_tuple = runtime_version_tuple(self._spark_version)
        if not version_tuple:
            return 0, 0
        return version_tuple

    @property
    def data_security_mode(self) -> compute.DataSecurityMode:
        return self._data_security_mode or compute.DataSecurityMode.NONE

    def build_dependency_graph(self, parent: DependencyGraph) -> list[DependencyProblem]:
        return list(self._register_task_dependencies(parent))

    def _register_task_dependencies(self, graph: DependencyGraph) -> Iterable[DependencyProblem]:
        yield from self._register_cluster_info()
        yield from self._register_libraries(graph)
        yield from self._register_existing_cluster_id(graph)
        yield from self._register_notebook(graph)
        yield from self._register_spark_python_task(graph)
        yield from self._register_python_wheel_task(graph)
        yield from self._register_spark_jar_task(graph)
        yield from self._register_run_job_task(graph)
        yield from self._register_pipeline_task(graph)
        yield from self._register_spark_submit_task(graph)

    def _register_libraries(self, graph: DependencyGraph) -> Iterable[DependencyProblem]:
        if not self._task.libraries:
            return
        for library in self._task.libraries:
            yield from self._register_library(graph, library)

    def _as_path(self, path: str) -> Path:
        parsed_path = parse.urlparse(path)
        match parsed_path.scheme:
            case "":
                return self._cache.get_path(path)
            case "dbfs":
                return DBFSPath(self._ws, parsed_path.path)
            case other:
                msg = f"Unsupported schema: {other} (only DBFS or Workspace paths are allowed)"
                raise ValueError(msg)

    @classmethod
    @contextmanager
    def _temporary_copy(cls, path: Path) -> Generator[Path, None, None]:
        with tempfile.TemporaryDirectory() as directory:
            temporary_path = Path(directory) / path.name
            with path.open("rb") as src, temporary_path.open("wb") as dst:
                shutil.copyfileobj(src, dst)
            yield temporary_path

    def _register_library(self, graph: DependencyGraph, library: compute.Library) -> Iterable[DependencyProblem]:
        if library.pypi:
            problems = graph.register_library(library.pypi.package)
            if problems:
                yield from problems
        if library.egg:
            yield from self._register_egg(graph, library)
        if library.whl:
            wheel_path = self._as_path(library.whl)
            with self._temporary_copy(wheel_path) as local_file:
                yield from graph.register_library(local_file.as_posix())
        if library.requirements:  # https://pip.pypa.io/en/stable/reference/requirements-file-format/
            logger.info(f"Registering libraries from {library.requirements}")
            requirements_path = self._as_path(library.requirements)
            with requirements_path.open() as requirements:
                for requirement in requirements:
                    requirement = requirement.rstrip()
                    clean_requirement = requirement.replace(" ", "")  # requirements.txt may contain spaces
                    if clean_requirement.startswith("-r"):
                        logger.warning(f"Reference to other requirements file is not supported: {requirement}")
                        continue
                    if clean_requirement.startswith("-c"):
                        logger.warning(f"Reference to constraints file is not supported: {requirement}")
                        continue
                    yield from graph.register_library(clean_requirement)
        if library.jar:
            # TODO: https://github.com/databrickslabs/ucx/issues/1641
            yield DependencyProblem('not-yet-implemented', 'Jar library is not yet implemented')

    def _register_egg(self, graph, library):
        if self.runtime_version > (14, 0):
            yield DependencyProblem(
                code='not-supported',
                message='Installing eggs is no longer supported on Databricks 14.0 or higher',
            )
        logger.info(f"Registering library from {library.egg}")
        egg_path = self._as_path(library.egg)
        with self._temporary_copy(egg_path) as local_file:
            yield from graph.register_library(local_file.as_posix())

    def _register_notebook(self, graph: DependencyGraph) -> Iterable[DependencyProblem]:
        if not self._task.notebook_task:
            return []
        self._named_parameters = self._task.notebook_task.base_parameters
        notebook_path = self._task.notebook_task.notebook_path
        logger.info(f'Discovering {self._task.task_key} entrypoint: {notebook_path}')
        # Notebooks can't be on DBFS.
        path = self._cache.get_path(notebook_path)
        return graph.register_notebook(path, False)

    def _register_spark_python_task(self, graph: DependencyGraph):
        if not self._task.spark_python_task:
            return []
        self._parameters = self._task.spark_python_task.parameters
        python_file = self._task.spark_python_task.python_file
        logger.info(f'Discovering {self._task.task_key} entrypoint: {python_file}')
        path = self._as_path(python_file)
        return graph.register_file(path)

    @staticmethod
    def _find_first_matching_distribution(path_lookup: PathLookup, name: str) -> metadata.Distribution | None:
        # Prepared exists in importlib.metadata.__init__pyi, but is not defined in importlib.metadata.__init__.py
        normalize_name = metadata.Prepared.normalize  # type: ignore
        normalized_name = normalize_name(name)
        for library_root in path_lookup.library_roots:
            for path in library_root.glob("*.dist-info"):
                distribution = metadata.Distribution.at(path)
                if normalize_name(distribution.name) == normalized_name:
                    return distribution
        return None

    def _register_python_wheel_task(self, graph: DependencyGraph) -> Iterable[DependencyProblem]:
        if not self._task.python_wheel_task:
            return []
        self._named_parameters = self._task.python_wheel_task.named_parameters
        self._parameters = self._task.python_wheel_task.parameters
        distribution_name = self._task.python_wheel_task.package_name
        distribution = self._find_first_matching_distribution(graph.path_lookup, distribution_name)
        if distribution is None:
            return [DependencyProblem("distribution-not-found", f"Could not find distribution for {distribution_name}")]
        entry_point_name = self._task.python_wheel_task.entry_point
        try:
            entry_point = distribution.entry_points[entry_point_name]
        except KeyError:
            return [
                DependencyProblem(
                    "distribution-entry-point-not-found",
                    f"Could not find distribution entry point for {distribution_name}.{entry_point_name}",
                )
            ]
        return graph.register_import(entry_point.module)

    def _register_spark_jar_task(self, graph: DependencyGraph):  # pylint: disable=unused-argument
        if not self._task.spark_jar_task:
            return
        # TODO: https://github.com/databrickslabs/ucx/issues/1641
        self._parameters = self._task.spark_jar_task.parameters
        yield DependencyProblem('not-yet-implemented', 'Spark Jar task is not yet implemented')

    def _register_run_job_task(self, graph: DependencyGraph):  # pylint: disable=unused-argument
        if not self._task.run_job_task:
            return
        # TODO: it's not clear how to terminate the graph
        yield DependencyProblem('not-yet-implemented', 'Run job task is not yet implemented')

    def _register_pipeline_task(self, graph: DependencyGraph):
        if not self._task.pipeline_task:
            return

        pipeline = self._ws.pipelines.get(self._task.pipeline_task.pipeline_id)
        if not pipeline.spec:
            return
        if not pipeline.spec.libraries:
            return

        pipeline_libraries = pipeline.spec.libraries
        for library in pipeline_libraries:
            if not library.notebook:
                return
            if library.notebook.path:
                notebook_path = library.notebook.path
                # Notebooks can't be on DBFS.
                path = self._cache.get_path(notebook_path)
                # the notebook is the root of the graph, so there's no context to inherit
                yield from graph.register_notebook(path, inherit_context=False)
            if library.jar:
                yield from self._register_library(graph, compute.Library(jar=library.jar))
            if library.maven:
                yield DependencyProblem('not-yet-implemented', 'Maven library is not yet implemented')
            if library.file:
                yield DependencyProblem('not-yet-implemented', 'File library is not yet implemented')

    def _register_existing_cluster_id(self, graph: DependencyGraph):
        if not self._task.existing_cluster_id:
            return

        # load libraries installed on the referred cluster
        library_full_status_list = self._ws.libraries.cluster_status(self._task.existing_cluster_id)

        for library_full_status in library_full_status_list:
            if library_full_status.library:
                yield from self._register_library(graph, library_full_status.library)

    def _register_spark_submit_task(self, graph: DependencyGraph):  # pylint: disable=unused-argument
        if not self._task.spark_submit_task:
            return
        yield DependencyProblem('not-yet-implemented', 'Spark submit task is not yet implemented')

    def _register_cluster_info(self):
        if self._task.existing_cluster_id:
            cluster_info = self._ws.clusters.get(self._task.existing_cluster_id)
            return self._new_job_cluster_metadata(cluster_info)
        if self._task.new_cluster:
            return self._new_job_cluster_metadata(self._task.new_cluster)
        if self._task.job_cluster_key:
            for job_cluster in self._job.settings.job_clusters:
                if job_cluster.job_cluster_key != self._task.job_cluster_key:
                    continue
                return self._new_job_cluster_metadata(job_cluster.new_cluster)
        self._data_security_mode = compute.DataSecurityMode.USER_ISOLATION
        self._is_serverless = True
        return []

    def _new_job_cluster_metadata(self, new_cluster):
        self._spark_conf = new_cluster.spark_conf
        self._spark_version = new_cluster.spark_version
        self._data_security_mode = new_cluster.data_security_mode
        return []


class WorkflowLinter:
    def __init__(
        self,
        ws: WorkspaceClient,
        resolver: DependencyResolver,
        path_lookup: PathLookup,
        migration_index: TableMigrationIndex,
        directfs_crawler: DirectFsAccessCrawler,
        include_job_ids: list[int] | None = None,
    ):
        self._ws = ws
        self._resolver = resolver
        self._path_lookup = path_lookup
        self._migration_index = migration_index
        self._directfs_crawler = directfs_crawler
        self._include_job_ids = include_job_ids

    def refresh_report(self, sql_backend: SqlBackend, inventory_database: str):
        tasks = []
        all_jobs = list(self._ws.jobs.list())
        logger.info(f"Preparing {len(all_jobs)} linting tasks...")
        for job in all_jobs:
            if self._include_job_ids and job.job_id not in self._include_job_ids:
                logger.info(f"Skipping job {job.job_id}...")
                continue
            tasks.append(functools.partial(self.lint_job, job.job_id))
        logger.info(f"Running {tasks} linting tasks in parallel...")
        job_results, errors = Threads.gather('linting workflows', tasks)
        job_problems: list[JobProblem] = []
        job_dfsas: list[DirectFsAccess] = []
        for problems, dfsas in job_results:
            job_problems.extend(problems)
            job_dfsas.extend(dfsas)
        logger.info(f"Saving {len(job_problems)} linting problems...")
        sql_backend.save_table(
            f'{inventory_database}.workflow_problems',
            job_problems,
            JobProblem,
            mode='overwrite',
        )
        self._directfs_crawler.dump_all(job_dfsas)
        if len(errors) > 0:
            raise ManyError(errors)

    def lint_job(self, job_id: int) -> tuple[list[JobProblem], list[DirectFsAccess]]:
        try:
            job = self._ws.jobs.get(job_id)
        except NotFound:
            logger.warning(f'Could not find job: {job_id}')
            return ([], [])

        problems, dfsas = self._lint_job(job)
        if len(problems) > 0:
            problem_messages = "\n".join([problem.as_message() for problem in problems])
            logger.warning(f"Found job problems:\n{problem_messages}")
        return problems, dfsas

    _UNKNOWN = Path('<UNKNOWN>')

    def _lint_job(self, job: jobs.Job) -> tuple[list[JobProblem], list[DirectFsAccess]]:
        problems: list[JobProblem] = []
        dfsas: list[DirectFsAccess] = []
        assert job.job_id is not None
        assert job.settings is not None
        assert job.settings.name is not None
        assert job.settings.tasks is not None
        linted_paths: set[Path] = set()
        for task in job.settings.tasks:
            graph, advices, session_state = self._build_task_dependency_graph(task, job)
            if not advices:
                advices = self._lint_task(task, graph, session_state, linted_paths)
            for advice in advices:
                absolute_path = advice.path.absolute().as_posix() if advice.path != self._UNKNOWN else 'UNKNOWN'
                job_problem = JobProblem(
                    job_id=job.job_id,
                    job_name=job.settings.name,
                    task_key=task.task_key,
                    path=absolute_path,
                    code=advice.advice.code,
                    message=advice.advice.message,
                    start_line=advice.advice.start_line,
                    start_col=advice.advice.start_col,
                    end_line=advice.advice.end_line,
                    end_col=advice.advice.end_col,
                )
                problems.append(job_problem)
            assessment_start = datetime.now(timezone.utc)
            task_dfsas = self._collect_task_dfsas(job, task, graph, session_state)
            assessment_end = datetime.now(timezone.utc)
            for dfsa in task_dfsas:
                dfsa = dfsa.replace_assessment_infos(assessment_start=assessment_start, assessment_end=assessment_end)
                dfsas.append(dfsa)
        return problems, dfsas

    def _build_task_dependency_graph(
        self, task: jobs.Task, job: jobs.Job
    ) -> tuple[DependencyGraph, Iterable[LocatedAdvice], CurrentSessionState]:
        root_dependency: Dependency = WorkflowTask(self._ws, task, job)
        # we can load it without further preparation since the WorkflowTask is merely a wrapper
        container = root_dependency.load(self._path_lookup)
        assert isinstance(container, WorkflowTaskContainer)
        session_state = CurrentSessionState(
            data_security_mode=container.data_security_mode,
            named_parameters=container.named_parameters,
            spark_conf=container.spark_conf,
            dbr_version=container.runtime_version,
        )
        graph = DependencyGraph(root_dependency, None, self._resolver, self._path_lookup, session_state)
        problems = container.build_dependency_graph(graph)
        located_advices: list[LocatedAdvice] = []
        for problem in problems:
            source_path = self._UNKNOWN if problem.is_path_missing() else problem.source_path
            located_advices.append(LocatedAdvice(problem.as_advisory(), source_path))
        return graph, located_advices, session_state

    def _lint_task(
        self,
        task: jobs.Task,
        graph: DependencyGraph,
        session_state: CurrentSessionState,
        linted_paths: set[Path],
    ) -> Iterable[LocatedAdvice]:
        walker = LintingWalker(
            graph, linted_paths, self._path_lookup, task.task_key, session_state, self._migration_index
        )
        yield from walker

    def _collect_task_dfsas(
        self, job: jobs.Job, task: jobs.Task, graph: DependencyGraph, session_state: CurrentSessionState
    ) -> Iterable[DirectFsAccess]:
        # walker doesn't register lineage for job/task
        job_id = str(job.job_id)
        job_name = job.settings.name if job.settings and job.settings.name else "<anonymous>"
        for dfsa in DfsaCollectorWalker(graph, set(), self._path_lookup, session_state):
            atoms = [
                LineageAtom(object_type="JOB", object_id=job_id, other={"name": job_name}),
                LineageAtom(object_type="TASK", object_id=task.task_key),
            ]
            yield dataclasses.replace(dfsa, source_lineage=atoms + dfsa.source_lineage)


class LintingWalker(DependencyGraphWalker[LocatedAdvice]):

    def __init__(
        self,
        graph: DependencyGraph,
        walked_paths: set[Path],
        path_lookup: PathLookup,
        key: str,
        session_state: CurrentSessionState,
        migration_index: TableMigrationIndex,
    ):
        super().__init__(graph, walked_paths, path_lookup)
        self._key = key
        self._session_state = session_state
        self._migration_index = migration_index

    def _log_walk_one(self, dependency: Dependency):
        logger.info(f'Linting {self._key} dependency: {dependency}')

    def _process_dependency(
        self, dependency: Dependency, path_lookup: PathLookup, inherited_tree: Tree | None
    ) -> Iterable[LocatedAdvice]:
        ctx = LinterContext(self._migration_index, self._session_state)
        # FileLinter determines which file/notebook linter to use
        linter = FileLinter(ctx, path_lookup, self._session_state, dependency.path, inherited_tree)
        for advice in linter.lint():
            yield LocatedAdvice(advice, dependency.path)


class DfsaCollectorWalker(DependencyGraphWalker[DirectFsAccess]):

    def __init__(
        self,
        graph: DependencyGraph,
        walked_paths: set[Path],
        path_lookup: PathLookup,
        session_state: CurrentSessionState,
    ):
        super().__init__(graph, walked_paths, path_lookup)
        self._session_state = session_state

    def _process_dependency(
        self, dependency: Dependency, path_lookup: PathLookup, inherited_tree: Tree | None
    ) -> Iterable[DirectFsAccess]:
        language = file_language(dependency.path)
        if not language:
            logger.warning(f"Unknown language for {dependency.path}")
            return
        cell_language = CellLanguage.of_language(language)
        source = dependency.path.read_text(guess_encoding(dependency.path))
        if is_a_notebook(dependency.path):
            yield from self._collect_from_notebook(source, cell_language, dependency.path, inherited_tree)
        elif dependency.path.is_file():
            yield from self._collect_from_source(source, cell_language, dependency.path, inherited_tree)

    def _collect_from_notebook(
        self, source: str, language: CellLanguage, path: Path, inherited_tree: Tree | None
    ) -> Iterable[DirectFsAccess]:
        notebook = Notebook.parse(path, source, language.language)
        src_timestamp = datetime.fromtimestamp(path.stat().st_mtime, timezone.utc)
        src_id = str(path)
        for cell in notebook.cells:
            for dfsa in self._collect_from_source(cell.original_code, cell.language, path, inherited_tree):
                yield dfsa.replace_source(source_id=src_id, source_lineage=self.lineage, source_timestamp=src_timestamp)
            if cell.language is CellLanguage.PYTHON:
                if inherited_tree is None:
                    inherited_tree = Tree.new_module()
                tree = Tree.normalize_and_parse(cell.original_code)
                inherited_tree.append_tree(tree)

    def _collect_from_source(
        self, source: str, language: CellLanguage, path: Path, inherited_tree: Tree | None
    ) -> Iterable[DirectFsAccess]:
        iterable: Iterable[DirectFsAccess] | None = None
        if language is CellLanguage.SQL:
            iterable = self._collect_from_sql(source)
        if language is CellLanguage.PYTHON:
            iterable = self._collect_from_python(source, inherited_tree)
        if iterable is None:
            logger.warning(f"Language {language.name} not supported yet!")
            return
        src_timestamp = datetime.fromtimestamp(path.stat().st_mtime, timezone.utc)
        src_id = str(path)
        for dfsa in iterable:
            yield dfsa.replace_source(source_id=src_id, source_lineage=self.lineage, source_timestamp=src_timestamp)

    def _collect_from_python(self, source: str, inherited_tree: Tree | None) -> Iterable[DirectFsAccess]:
        linter = DirectFsAccessPyLinter(self._session_state, prevent_spark_duplicates=False)
        for dfsa_node in linter.collect_dfsas(source, inherited_tree):
            yield dfsa_node.dfsa

    def _collect_from_sql(self, source: str) -> Iterable[DirectFsAccess]:
        linter = DirectFsAccessSqlLinter()
        yield from linter.collect_dfsas(source)



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from pathlib import Path
from typing import TypeVar

from astroid import NodeNG, Call, Name, Const  # type: ignore

from databricks.labs.ucx.source_code.graph import (
    DependencyGraph,
    DependencyProblem,
    DependencyGraphContext,
    InheritedContext,
)
from databricks.labs.ucx.source_code.python.python_ast import NodeBase, Tree


logger = logging.getLogger(__name__)

T = TypeVar("T")


class MagicLine(NodeBase):

    @classmethod
    def extract_from_tree(
        cls, tree: Tree, problem_factory: Callable[[str, str, NodeNG], T]
    ) -> tuple[list[MagicLine], list[T]]:
        problems: list[T] = []
        commands: list[MagicLine] = []
        try:
            nodes = tree.locate(Call, [("magic_command", Name)])
            for command in cls._make_commands_for_magic_command_call_nodes(nodes):
                commands.append(command)
        except Exception as e:  # pylint: disable=broad-except
            logger.debug(f"Internal error while checking magic commands in tree: {tree.root}", exc_info=True)
            problem = problem_factory('internal-error', f"While checking magic commands: {e}", tree.root)
            problems.append(problem)
        return commands, problems

    @classmethod
    def _make_commands_for_magic_command_call_nodes(cls, nodes: list[Call]):
        for node in nodes:
            arg = node.args[0]
            if isinstance(arg, Const):
                yield MagicLine(node, arg.value)

    def __init__(self, node: NodeNG, command: bytes):
        super().__init__(node)
        self._command = command.decode()

    def as_magic(self) -> MagicCommand | None:
        for factory in _FACTORIES:
            command = factory(self._command, self.node)
            if command is not None:
                return command
        return None

    def build_dependency_graph(self, parent: DependencyGraph) -> list[DependencyProblem]:
        magic = self.as_magic()
        if magic is not None:
            return magic.build_dependency_graph(parent)
        problem = DependencyProblem.from_node(
            code='unsupported-magic-line', message=f"magic line '{self._command}' is not supported yet", node=self.node
        )
        return [problem]

    def build_inherited_context(self, context: DependencyGraphContext, child_path: Path) -> InheritedContext:
        magic = self.as_magic()
        if magic is not None:
            return magic.build_inherited_context(context, child_path)
        return InheritedContext(None, False)


class MagicNode(NodeNG):
    pass


class MagicCommand(ABC):

    def __init__(self, node: NodeNG, code: str):
        self._node = node
        self._code = code

    @abstractmethod
    def build_dependency_graph(self, parent: DependencyGraph) -> list[DependencyProblem]: ...

    def build_inherited_context(self, _context: DependencyGraphContext, _child_path: Path) -> InheritedContext:
        return InheritedContext(None, False)


_FACTORIES: list[Callable[[str, NodeNG], MagicCommand | None]] = []


def magic_command_factory(func: Callable[[str, NodeNG], MagicCommand | None]):
    _FACTORIES.append(func)

    def inner(command: str, node: NodeNG) -> MagicCommand | None:
        return func(command, node)

    return inner



import base64
import json
import logging
from collections.abc import Iterable
from dataclasses import dataclass

from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound, BadRequest
from databricks.sdk.service.compute import (
    ClusterDetails,
    ClusterSource,
    DataSecurityMode,
    DbfsStorageInfo,
    InitScriptInfo,
    LocalFileInfo,
    Policy,
    WorkspaceStorageInfo,
)

from databricks.labs.ucx.assessment.crawlers import (
    AZURE_SP_CONF_FAILURE_MSG,
    INCOMPATIBLE_SPARK_CONFIG_KEYS,
    INIT_SCRIPT_DBFS_PATH,
    INIT_SCRIPT_LOCAL_PATH,
    azure_sp_conf_present_check,
    spark_version_compatibility,
    is_mlr,
)
from databricks.labs.ucx.assessment.init_scripts import CheckInitScriptMixin
from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier

logger = logging.getLogger(__name__)


@dataclass
class ClusterInfo:
    cluster_id: str
    success: int
    failures: str
    spark_version: str | None = None
    policy_id: str | None = None
    cluster_name: str | None = None
    creator: str | None = None


class CheckClusterMixin(CheckInitScriptMixin):
    _ws: WorkspaceClient

    def _safe_get_cluster_policy(self, policy_id: str) -> Policy | None:
        try:
            return self._ws.cluster_policies.get(policy_id)
        except NotFound:
            logger.warning(f"The cluster policy was deleted: {policy_id}")
            return None

    def _check_cluster_policy(self, policy_id: str, source: str) -> list[str]:
        failures: list[str] = []
        policy = self._safe_get_cluster_policy(policy_id)
        if policy:
            if policy.definition:
                if azure_sp_conf_present_check(json.loads(policy.definition)):
                    failures.append(f"{AZURE_SP_CONF_FAILURE_MSG} {source}.")
            if policy.policy_family_definition_overrides:
                if azure_sp_conf_present_check(json.loads(policy.policy_family_definition_overrides)):
                    failures.append(f"{AZURE_SP_CONF_FAILURE_MSG} {source}.")
        return failures

    def _get_init_script_data(self, init_script_info: InitScriptInfo) -> str | None:
        try:
            match init_script_info:
                case InitScriptInfo(dbfs=DbfsStorageInfo(destination)):
                    split = destination.split(":")
                    if len(split) != INIT_SCRIPT_DBFS_PATH:
                        return None
                    data = self._ws.dbfs.read(split[1]).data
                    if data is not None:
                        return base64.b64decode(data).decode("utf-8")
                case InitScriptInfo(workspace=WorkspaceStorageInfo(workspace_file_destination)):
                    data = self._ws.workspace.export(workspace_file_destination).content
                    if data is not None:
                        return base64.b64decode(data).decode("utf-8")
                case InitScriptInfo(file=LocalFileInfo(destination)):
                    split = destination.split(":/")
                    if len(split) != INIT_SCRIPT_LOCAL_PATH:
                        return None
                    with open(split[1], "r", encoding="utf-8") as file:
                        data = file.read()
                    return data

            return None
        except (NotFound, BadRequest):
            return None

    def _check_cluster_init_script(self, init_scripts: list[InitScriptInfo], source: str) -> list[str]:
        failures: list[str] = []
        for init_script_info in init_scripts:
            init_script_data = self._get_init_script_data(init_script_info)
            failures.extend(self.check_init_script(init_script_data, source))
        return failures

    def _check_spark_conf(self, conf: dict[str, str], source: str) -> list[str]:
        failures: list[str] = []
        for key, error in INCOMPATIBLE_SPARK_CONFIG_KEYS.items():
            if key in conf:
                failures.append(f"{error}: {key} in {source}.")
        for value in conf.values():
            if "dbfs:/mnt" in value or "/dbfs/mnt" in value:
                failures.append(f"using DBFS mount in configuration: {value}")
        # Checking if Azure cluster config is present in spark config
        if azure_sp_conf_present_check(conf):
            failures.append(f"{AZURE_SP_CONF_FAILURE_MSG} {source}.")
        return failures

    def _check_cluster_failures(self, cluster: ClusterDetails, source: str) -> list[str]:
        failures: list[str] = []

        unsupported_cluster_types = [
            DataSecurityMode.LEGACY_PASSTHROUGH,
            DataSecurityMode.LEGACY_SINGLE_USER,
            DataSecurityMode.LEGACY_TABLE_ACL,
        ]
        support_status = spark_version_compatibility(cluster.spark_version)
        if support_status != "supported":
            failures.append(f"not supported DBR: {cluster.spark_version}")
        if cluster.spark_conf is not None:
            failures.extend(self._check_spark_conf(cluster.spark_conf, source))
        # Checking if Azure cluster config is present in cluster policies
        if cluster.policy_id is not None:
            failures.extend(self._check_cluster_policy(cluster.policy_id, source))
        if cluster.init_scripts is not None:
            failures.extend(self._check_cluster_init_script(cluster.init_scripts, source))
        if cluster.data_security_mode == DataSecurityMode.NONE:
            failures.append("No isolation shared clusters not supported in UC")
        if cluster.data_security_mode in unsupported_cluster_types:
            failures.append(f"cluster type not supported : {cluster.data_security_mode.value}")
        if cluster.data_security_mode == DataSecurityMode.NONE and is_mlr(cluster.spark_version):
            failures.append("Shared Machine Learning Runtime clusters are not supported in UC")

        return failures


class ClustersCrawler(CrawlerBase[ClusterInfo], CheckClusterMixin):
    def __init__(self, ws: WorkspaceClient, sbe: SqlBackend, schema: str):
        super().__init__(sbe, "hive_metastore", schema, "clusters", ClusterInfo)
        self._ws = ws

    def _crawl(self) -> Iterable[ClusterInfo]:
        all_clusters = list(self._ws.clusters.list())
        return list(self._assess_clusters(all_clusters))

    def _assess_clusters(self, all_clusters):
        for cluster in all_clusters:
            if cluster.cluster_source == ClusterSource.JOB:
                continue
            if not cluster.creator_user_name:
                logger.warning(
                    f"Cluster {cluster.cluster_id} have Unknown creator, it means that the original creator "
                    f"has been deleted and should be re-created"
                )
            cluster_info = ClusterInfo(
                cluster_id=cluster.cluster_id if cluster.cluster_id else "",
                cluster_name=cluster.cluster_name,
                policy_id=cluster.policy_id,
                spark_version=cluster.spark_version,
                creator=cluster.creator_user_name,
                success=1,
                failures="[]",
            )
            failures = self._check_cluster_failures(cluster, "cluster")
            if len(failures) > 0:
                cluster_info.success = 0
                cluster_info.failures = json.dumps(failures)
            yield cluster_info

    def _try_fetch(self) -> Iterable[ClusterInfo]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield ClusterInfo(*row)


@dataclass
class PolicyInfo:
    policy_id: str
    policy_name: str
    success: int
    failures: str
    spark_version: str | None = None
    policy_description: str | None = None
    creator: str | None = None


class PoliciesCrawler(CrawlerBase[PolicyInfo], CheckClusterMixin):
    def __init__(self, ws: WorkspaceClient, sbe: SqlBackend, schema):
        super().__init__(sbe, "hive_metastore", schema, "policies", PolicyInfo)
        self._ws = ws

    def _crawl(self) -> Iterable[PolicyInfo]:
        all_policices = list(self._ws.cluster_policies.list())
        return list(self._assess_policies(all_policices))

    def _assess_policies(self, all_policices) -> Iterable[PolicyInfo]:
        for policy in all_policices:
            failures: list[str] = []
            if policy.policy_id is None:
                continue
            failures.extend(self._check_cluster_policy(policy.policy_id, "policy"))
            try:
                spark_version = json.dumps(json.loads(policy.definition)["spark_version"])
            except KeyError:
                spark_version = None
            policy_name = policy.name
            creator_name = policy.creator_user_name

            policy_info = PolicyInfo(
                policy_id=policy.policy_id,
                policy_description=policy.description,
                policy_name=policy_name,
                spark_version=spark_version,
                success=1,
                failures="[]",
                creator=creator_name,
            )
            if len(failures) > 0:
                policy_info.success = 0
                policy_info.failures = json.dumps(failures)
            yield policy_info

    def _try_fetch(self) -> Iterable[PolicyInfo]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield PolicyInfo(*row)



from __future__ import annotations

import logging
from collections.abc import Iterable, Iterator, Generator
from typing import Any

from astroid import (  # type: ignore
    Attribute,
    Call,
    Const,
    decorators,
    Dict,
    FormattedValue,
    JoinedStr,
    Name,
    NodeNG,
    Uninferable,
)
from astroid.context import InferenceContext, InferenceResult, CallContext  # type: ignore
from astroid.typing import InferenceErrorInfo  # type: ignore
from astroid.exceptions import InferenceError  # type: ignore

from databricks.labs.ucx.source_code.base import CurrentSessionState
from databricks.labs.ucx.source_code.python.python_ast import Tree

logger = logging.getLogger(__name__)


class InferredValue:
    """Represents 1 or more nodes that together represent the value.
    The list of nodes typically holds one Const element, but for f-strings it
    can hold multiple ones, including Uninferable nodes."""

    @classmethod
    def infer_from_node(cls, node: NodeNG, state: CurrentSessionState | None = None) -> Iterable[InferredValue]:
        cls._contextualize(node, state)
        for inferred_atoms in cls._infer_values(node):
            yield InferredValue(inferred_atoms)

    @classmethod
    def _contextualize(cls, node: NodeNG, state: CurrentSessionState | None):
        if state is None or state.named_parameters is None or len(state.named_parameters) == 0:
            return
        cls._contextualize_dbutils_widgets_get(node, state)
        cls._contextualize_dbutils_widgets_get_all(node, state)

    @classmethod
    def _contextualize_dbutils_widgets_get(cls, node: NodeNG, state: CurrentSessionState):
        root = Tree(node).root
        calls = Tree(root).locate(Call, [("get", Attribute), ("widgets", Attribute), ("dbutils", Name)])
        for call in calls:
            call.func = _DbUtilsWidgetsGetCall(state, call)

    @classmethod
    def _contextualize_dbutils_widgets_get_all(cls, node: NodeNG, state: CurrentSessionState):
        root = Tree(node).root
        calls = Tree(root).locate(Call, [("getAll", Attribute), ("widgets", Attribute), ("dbutils", Name)])
        for call in calls:
            call.func = _DbUtilsWidgetsGetAllCall(state, call)

    @classmethod
    def _infer_values(cls, node: NodeNG) -> Iterator[Iterable[NodeNG]]:
        # deal with node types that don't implement 'inferred()'
        if node is Uninferable or isinstance(node, Const):
            yield [node]
        elif isinstance(node, JoinedStr):
            yield from cls._infer_values_from_joined_string(node)
        elif isinstance(node, FormattedValue):
            yield from _LocalInferredValue.do_infer_values(node.value)
        else:
            yield from cls._infer_internal(node)

    @classmethod
    def _infer_internal(cls, node: NodeNG):
        try:
            for inferred in node.inferred():
                # work around infinite recursion of empty lists
                if inferred == node:
                    continue
                yield from _LocalInferredValue.do_infer_values(inferred)
        except InferenceError as e:
            logger.debug(f"When inferring {node}", exc_info=e)
            yield [Uninferable]

    @classmethod
    def _infer_values_from_joined_string(cls, node: NodeNG) -> Iterator[Iterable[NodeNG]]:
        assert isinstance(node, JoinedStr)
        yield from cls._infer_values_from_joined_values(node.values)

    @classmethod
    def _infer_values_from_joined_values(cls, nodes: list[NodeNG]) -> Iterator[Iterable[NodeNG]]:
        if len(nodes) == 1:
            yield from _LocalInferredValue.do_infer_values(nodes[0])
            return
        for firsts in _LocalInferredValue.do_infer_values(nodes[0]):
            for remains in cls._infer_values_from_joined_values(nodes[1:]):
                yield list(firsts) + list(remains)

    def __init__(self, atoms: Iterable[NodeNG]):
        self._atoms = list(atoms)

    @property
    def nodes(self):
        return self._atoms

    def is_inferred(self):
        return all(atom is not Uninferable for atom in self._atoms)

    def as_string(self):
        strings = [str(const.value) for const in filter(lambda atom: isinstance(atom, Const), self._atoms)]
        return "".join(strings)


class _DbUtilsWidgetsGetCall(NodeNG):

    def __init__(self, session_state: CurrentSessionState, node: NodeNG):
        super().__init__(
            lineno=node.lineno,
            col_offset=node.col_offset,
            end_lineno=node.end_lineno,
            end_col_offset=node.end_col_offset,
            parent=node.parent,
        )
        self._session_state = session_state

    @decorators.raise_if_nothing_inferred
    def _infer(
        self, context: InferenceContext | None = None, **kwargs: Any
    ) -> Generator[InferenceResult, None, InferenceErrorInfo | None]:
        yield self
        return InferenceErrorInfo(node=self, context=context)

    def infer_call_result(self, context: InferenceContext | None = None, **_):  # caller needs unused kwargs
        call_context = getattr(context, "callcontext", None)
        if not isinstance(call_context, CallContext):
            yield Uninferable
            return
        arg = call_context.args[0]
        for inferred in InferredValue.infer_from_node(arg, self._session_state):
            if not inferred.is_inferred():
                yield Uninferable
                continue
            name = inferred.as_string()
            named_parameters = self._session_state.named_parameters
            if not named_parameters or name not in named_parameters:
                yield Uninferable
                continue
            value = named_parameters[name]
            yield Const(
                value,
                lineno=self.lineno,
                col_offset=self.col_offset,
                end_lineno=self.end_lineno,
                end_col_offset=self.end_col_offset,
                parent=self,
            )


class _LocalInferredValue(InferredValue):

    @classmethod
    def do_infer_values(cls, node: NodeNG):
        yield from cls._infer_values(node)


class _DbUtilsWidgetsGetAllCall(NodeNG):

    def __init__(self, session_state: CurrentSessionState, node: NodeNG):
        super().__init__(
            lineno=node.lineno,
            col_offset=node.col_offset,
            end_lineno=node.end_lineno,
            end_col_offset=node.end_col_offset,
            parent=node.parent,
        )
        self._session_state = session_state

    @decorators.raise_if_nothing_inferred
    def _infer(
        self, context: InferenceContext | None = None, **kwargs: Any
    ) -> Generator[InferenceResult, None, InferenceErrorInfo | None]:
        yield self
        return InferenceErrorInfo(node=self, context=context)

    def infer_call_result(self, **_):  # caller needs unused kwargs
        named_parameters = self._session_state.named_parameters
        if not named_parameters:
            yield Uninferable
            return
        items = self._populate_items(named_parameters)
        result = Dict(
            lineno=self.lineno,
            col_offset=self.col_offset,
            end_lineno=self.end_lineno,
            end_col_offset=self.end_col_offset,
            parent=self,
        )
        result.postinit(items)
        yield result

    def _populate_items(self, values: dict[str, str]):
        items: list[tuple[InferenceResult, InferenceResult]] = []
        for key, value in values.items():
            item_key = Const(
                key,
                lineno=self.lineno,
                col_offset=self.col_offset,
                end_lineno=self.end_lineno,
                end_col_offset=self.end_col_offset,
                parent=self,
            )
            item_value = Const(
                value,
                lineno=self.lineno,
                col_offset=self.col_offset,
                end_lineno=self.end_lineno,
                end_col_offset=self.end_col_offset,
                parent=self,
            )
            items.append((item_key, item_value))
        return items



import json
import logging
import re
import uuid
from collections.abc import Callable, ValuesView
from dataclasses import dataclass
from functools import partial, wraps
from datetime import timedelta
from typing import Any, ParamSpec, TypeVar

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.parallel import ManyError, Threads
from databricks.labs.blueprint.tui import Prompts
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import (
    BadRequest,
    InvalidParameterValue,
    NotFound,
    PermissionDenied,
    ResourceAlreadyExists,
    ResourceDoesNotExist,
)
from databricks.sdk.retries import retried
from databricks.sdk.service.catalog import Privilege
from databricks.sdk.service.compute import Policy
from databricks.sdk.service.sql import (
    EndpointConfPair,
    GetWorkspaceWarehouseConfigResponse,
    SetWorkspaceWarehouseConfigRequestSecurityPolicy,
)
from databricks.sdk.service.workspace import GetSecretResponse

from databricks.labs.ucx.azure.resources import (
    AccessConnector,
    AzureResources,
    AzureRoleAssignment,
    PrincipalSecret,
    StorageAccount,
)
from databricks.labs.ucx.config import WorkspaceConfig
from databricks.labs.ucx.hive_metastore.locations import ExternalLocations

logger = logging.getLogger(__name__)
P = ParamSpec('P')
R = TypeVar('R')


@dataclass
class StoragePermissionMapping:
    prefix: str
    client_id: str
    principal: str
    privilege: str
    type: str
    default_network_action: str = "Unknown"  # "Deny", "Allow" or "Unknown"
    # Need this directory_id/tenant_id when create UC storage credentials using service principal
    directory_id: str | None = None


class AzureResourcePermissions:
    FILENAME = 'azure_storage_account_info.csv'
    _UBER_PRINCIPAL_SECRET_KEY = "uber_principal_secret"

    def __init__(
        self,
        installation: Installation,
        ws: WorkspaceClient,
        azurerm: AzureResources,
        external_locations: ExternalLocations,
    ):
        self._installation = installation
        self._locations = external_locations
        self._azurerm = azurerm
        self._ws = ws
        self._levels = {
            "Storage Blob Data Contributor": Privilege.WRITE_FILES,
            "Storage Blob Data Owner": Privilege.WRITE_FILES,
            "Storage Blob Data Reader": Privilege.READ_FILES,
        }
        self._permission_levels = {
            "Microsoft.Storage/storageAccounts/blobServices/containers/write": Privilege.WRITE_FILES,
            "Microsoft.Storage/storageAccounts/blobServices/containers/blobs/write": Privilege.WRITE_FILES,
            "Microsoft.Storage/storageAccounts/blobServices/containers/read": Privilege.READ_FILES,
            "Microsoft.Storage/storageAccounts/blobServices/containers/blobs/read": Privilege.READ_FILES,
        }

    def _get_permission_level(self, permission_to_match: str) -> Privilege | None:
        # String might contain '*', check for wildcard match
        pattern = re.sub(r'\*', '.*', permission_to_match)
        permission_compiled = re.compile(pattern)
        for each_level, privilege_level in self._permission_levels.items():
            # Check for storage blob permission with regex to account for star pattern
            match = permission_compiled.match(each_level)
            # If a match is found, return the privilege level, no need to check for lower levels
            if match:
                return privilege_level
        return None

    def _get_custom_role_privilege(self, role_permissions: list[str]) -> Privilege | None:
        # If both read and write privileges are found, only write privilege will be considered
        higher_privilege = None
        for each_permission in role_permissions:
            privilege = self._get_permission_level(each_permission)
            if privilege is None:
                continue
            # WRITE_FILES is the higher permission, don't need to check further
            if privilege == Privilege.WRITE_FILES:
                return privilege
            if privilege == Privilege.READ_FILES:
                higher_privilege = privilege
        return higher_privilege

    def _get_role_privilege(self, role_assignment: AzureRoleAssignment) -> Privilege | None:
        privilege = None
        # Check for custom role permissions on the storage accounts
        if role_assignment.role_permissions:
            privilege = self._get_custom_role_privilege(role_assignment.role_permissions)
        elif role_assignment.role_name in self._levels:
            privilege = self._levels[role_assignment.role_name]
        return privilege

    def _map_storage(self, storage: StorageAccount) -> ValuesView[StoragePermissionMapping]:
        logger.info(f"Fetching role assignment for {storage.name}")
        principal_spm_mapping: dict[str, StoragePermissionMapping] = {}
        for container in self._azurerm.containers(storage.id):
            for role_assignment in self._azurerm.role_assignments(str(container)):
                # Skip the role assignments that already have WRITE_FILES privilege
                spm_mapping_key = f"{container.container}_{role_assignment.principal.client_id}"
                if (
                    spm_mapping_key in principal_spm_mapping
                    and principal_spm_mapping[spm_mapping_key].privilege == Privilege.WRITE_FILES.value
                ):
                    continue
                # one principal may be assigned multiple roles with overlapping dataActions, hence appearing
                # here in duplicates. hence, role name -> permission level is not enough for the perfect scenario.
                returned_privilege = self._get_role_privilege(role_assignment)
                if not returned_privilege:
                    continue
                privilege = returned_privilege.value
                principal_spm_mapping[spm_mapping_key] = StoragePermissionMapping(
                    prefix=f"abfss://{container.container}@{container.storage_account}.dfs.core.windows.net/",
                    client_id=role_assignment.principal.client_id,
                    principal=role_assignment.principal.display_name,
                    privilege=privilege,
                    type=role_assignment.principal.type,
                    default_network_action=storage.default_network_action,
                    directory_id=role_assignment.principal.directory_id,
                )
        return principal_spm_mapping.values()

    def save_spn_permissions(self) -> str | None:
        storage_accounts = self._get_storage_accounts()
        if len(storage_accounts) == 0:
            logger.warning(
                "There are no external table present with azure storage account. "
                "Please check if assessment job is run"
            )
            return None
        storage_account_infos = []
        for storage in storage_accounts:
            for mapping in self._map_storage(storage):
                storage_account_infos.append(mapping)
        if len(storage_account_infos) == 0:
            logger.error("No storage account found in current tenant with spn permission")
            return None
        return self._installation.save(storage_account_infos, filename=self.FILENAME)

    @staticmethod
    def _policy_config(value: str) -> dict[str, str]:
        return {"type": "fixed", "value": value}

    def _create_service_principal_cluster_policy_configuration_pairs(
        self, principal_client_id: str, principal_secret_identifier: str, storage: StorageAccount
    ) -> list[tuple[str, dict[str, str]]]:
        """Create the cluster policy configuration pairs to access the storage"""
        tenant_id = self._azurerm.tenant_id()
        endpoint = f"https://login.microsoftonline.com/{tenant_id}/oauth2/token"
        configuration_pairs = [
            (
                f"spark_conf.fs.azure.account.oauth2.client.id.{storage.name}.dfs.core.windows.net",
                self._policy_config(principal_client_id),
            ),
            (
                f"spark_conf.fs.azure.account.oauth.provider.type.{storage.name}.dfs.core.windows.net",
                self._policy_config("org.apache.hadoop.fs.azurebfs.oauth2.ClientCredsTokenProvider"),
            ),
            (
                f"spark_conf.fs.azure.account.oauth2.client.endpoint.{storage.name}.dfs.core.windows.net",
                self._policy_config(endpoint),
            ),
            (
                f"spark_conf.fs.azure.account.auth.type.{storage.name}.dfs.core.windows.net",
                self._policy_config("OAuth"),
            ),
            (
                f"spark_conf.fs.azure.account.oauth2.client.secret.{storage.name}.dfs.core.windows.net",
                self._policy_config("{{" + principal_secret_identifier + "}}"),
            ),
        ]
        return configuration_pairs

    def _add_service_principal_configuration_to_cluster_policy_definition(
        self,
        policy_definition: str,
        principal_client_id: str,
        principal_secret_identifier: str,
        storage_accounts: list[StorageAccount],
    ) -> str:
        policy_dict = json.loads(policy_definition)
        for storage in storage_accounts:
            for key, value in self._create_service_principal_cluster_policy_configuration_pairs(
                principal_client_id, principal_secret_identifier, storage
            ):
                policy_dict[key] = value
        return json.dumps(policy_dict)

    def _add_service_principal_configuration_to_cluster_policy(
        self,
        policy_id: str,
        principal_client_id: str,
        principal_secret_identifier: str,
        storage_accounts: list[StorageAccount],
    ):
        policy_definition = ""
        try:
            cluster_policy = self._ws.cluster_policies.get(policy_id)
            self._installation.save(cluster_policy, filename="policy-backup.json")
            if cluster_policy.definition is not None:
                policy_definition = self._add_service_principal_configuration_to_cluster_policy_definition(
                    cluster_policy.definition,
                    principal_client_id,
                    principal_secret_identifier,
                    storage_accounts,
                )
            if cluster_policy.name is not None:
                self._ws.cluster_policies.edit(policy_id, name=cluster_policy.name, definition=policy_definition)
            logger.info(
                f"Updated UCX cluster policy {policy_id} with service principal connection details for accesing storage accounts"
            )
        except NotFound:
            msg = f"cluster policy {policy_id} not found, please run UCX installation to create UCX cluster policy"
            raise NotFound(msg) from None

    def _remove_service_principal_configuration_from_cluster_policy_definition(
        self,
        policy_definition: str,
        principal_client_id: str,
        principal_secret_identifier: str,
        storage_accounts: list[StorageAccount],
    ) -> str:
        policy_dict = json.loads(policy_definition)
        for storage in storage_accounts:
            for key, _ in self._create_service_principal_cluster_policy_configuration_pairs(
                principal_client_id, principal_secret_identifier, storage
            ):
                if key in policy_dict:
                    del policy_dict[key]
        return json.dumps(policy_dict)

    def _remove_service_principal_configuration_from_cluster_policy(
        self,
        policy_id: str,
        principal_client_id: str,
        principal_secret_identifier: str,
        storage_accounts: list[StorageAccount],
    ):
        """Revert the cluster policy."""
        try:
            policy = self._installation.load(Policy, filename="policy-backup.json")
        except NotFound:
            try:
                policy = self._ws.cluster_policies.get(policy_id)
            except (InvalidParameterValue, NotFound):
                return  # No policy to revert
        policy_definition = policy.definition
        if policy_definition is not None:
            policy_definition = self._remove_service_principal_configuration_from_cluster_policy_definition(
                policy_definition,
                principal_client_id,
                principal_secret_identifier,
                storage_accounts,
            )
        if policy.name is not None:
            self._ws.cluster_policies.edit(policy_id, name=policy.name, definition=policy_definition)

    def _create_storage_account_data_access_configuration_pairs(
        self, principal_client_id: str, principal_secret_identifier: str, storage: StorageAccount
    ) -> list[EndpointConfPair]:
        """Create the data access configuration pairs to access the storage"""
        configuration_pairs = []
        for key, value in self._create_service_principal_cluster_policy_configuration_pairs(
            principal_client_id, principal_secret_identifier, storage
        ):
            configuration_pairs.append(EndpointConfPair(key, value["value"]))
        return configuration_pairs

    def _add_service_principal_configuration_to_workspace_warehouse_config(
        self,
        principal_client_id: str,
        principal_secret_identifier: str,
        storage_accounts: list[StorageAccount],
    ):
        warehouse_config = self._ws.warehouses.get_workspace_warehouse_config()
        self._installation.save(warehouse_config, filename="warehouse-config-backup.json")
        sql_dac = warehouse_config.data_access_config or []
        for storage in storage_accounts:
            configuration_pairs = self._create_storage_account_data_access_configuration_pairs(
                principal_client_id, principal_secret_identifier, storage
            )
            sql_dac.extend(configuration_pairs)
        security_policy = (
            SetWorkspaceWarehouseConfigRequestSecurityPolicy(warehouse_config.security_policy.value)
            if warehouse_config.security_policy
            else SetWorkspaceWarehouseConfigRequestSecurityPolicy.NONE
        )
        try:
            self._ws.warehouses.set_workspace_warehouse_config(
                data_access_config=sql_dac,
                sql_configuration_parameters=warehouse_config.sql_configuration_parameters,
                security_policy=security_policy,
            )
            logger.info(
                "Updated workspace warehouse config with service principal connection details for accessing storage accounts"
            )
        # TODO: Remove following try except once https://github.com/databricks/databricks-sdk-py/issues/305 is fixed
        except InvalidParameterValue as error:
            sql_dac_log_msg = "\n".join(f"{config_pair.key} {config_pair.value}" for config_pair in sql_dac)
            logger.error(
                f'Adding uber principal to SQL warehouse Data Access Properties is failed using Python SDK with error "{error}". '
                f'Please try applying the following configs manually in the worksapce admin UI:\n{sql_dac_log_msg}'
            )
            raise error

    def _remove_service_principal_configuration_from_workspace_warehouse_config(
        self,
        principal_client_id: str,
        principal_secret_identifier: str,
        storage_accounts: list[StorageAccount],
    ):
        try:
            warehouse_config = self._installation.load(
                GetWorkspaceWarehouseConfigResponse, filename="warehouse-config-backup.json"
            )
        except NotFound:  # For legacy reasons we can not assume the backup to always be present
            warehouse_config = self._ws.warehouses.get_workspace_warehouse_config()
        sql_dac = warehouse_config.data_access_config or []

        for storage_account in storage_accounts:
            configuration_pairs = self._create_storage_account_data_access_configuration_pairs(
                principal_client_id, principal_secret_identifier, storage_account
            )
            for configuration_pair in configuration_pairs:
                if configuration_pair in sql_dac:
                    sql_dac.remove(configuration_pair)

        security_policy = (
            SetWorkspaceWarehouseConfigRequestSecurityPolicy(warehouse_config.security_policy.value)
            if warehouse_config.security_policy
            else SetWorkspaceWarehouseConfigRequestSecurityPolicy.NONE
        )
        try:
            self._ws.warehouses.set_workspace_warehouse_config(
                data_access_config=sql_dac,
                sql_configuration_parameters=warehouse_config.sql_configuration_parameters,
                security_policy=security_policy,
            )
        # TODO: Remove following try except once https://github.com/databricks/databricks-sdk-py/issues/305 is fixed
        except InvalidParameterValue as error:
            sql_dac_log_msg = "\n".join(f"{config_pair.key} {config_pair.value}" for config_pair in sql_dac)
            logger.error(
                f'Adding uber principal to SQL warehouse Data Access Properties is failed using Python SDK with error "{error}". '
                f'Please try applying the following configs manually in the workspace admin UI:\n{sql_dac_log_msg}'
            )
            raise error

    def create_uber_principal(self, prompts: Prompts) -> None:
        config = self._installation.load(WorkspaceConfig)
        inventory_database = config.inventory_database
        display_name = f"unity-catalog-migration-{inventory_database}-{self._ws.get_workspace_id()}"
        uber_principal_name = prompts.question(
            "Enter a name for the uber service principal to be created", default=display_name
        )
        policy_id = config.policy_id
        if policy_id is None:
            msg = "UCX cluster policy not found in config. Please run latest UCX installation to set cluster policy"
            logger.error(msg)
            raise ValueError(msg) from None
        if config.uber_spn_id is not None:
            logger.warning("Uber service principal already created for this workspace.")
            return
        storage_accounts = self._get_storage_accounts()
        if len(storage_accounts) == 0:
            logger.warning(
                "There are no external table present with azure storage account. "
                "Please check if assessment job is run"
            )
            return
        logger.info("Creating service principal")

        secret_identifier = f"secrets/{inventory_database}/{self._UBER_PRINCIPAL_SECRET_KEY}"
        try:
            uber_principal = self._azurerm.create_service_principal(uber_principal_name)
            config.uber_spn_id = uber_principal.client.client_id
            self._installation.save(config)
            self._apply_storage_permission(
                uber_principal.client.object_id, "STORAGE_BLOB_DATA_CONTRIBUTOR", *storage_accounts
            )
            self._create_and_get_secret_for_uber_principal(uber_principal, inventory_database)
            self._add_service_principal_configuration_to_cluster_policy(
                policy_id,
                uber_principal.client.client_id,
                secret_identifier,
                storage_accounts,
            )
            self._add_service_principal_configuration_to_workspace_warehouse_config(
                uber_principal.client.client_id, secret_identifier, storage_accounts
            )
        except (PermissionDenied, NotFound, BadRequest):
            self._delete_uber_principal()  # Clean up dangling resources
            raise

    def _delete_uber_principal(self) -> None:

        def log_permission_denied(function: Callable[P, R], *, message: str) -> Callable[P, R | None]:
            @wraps(function)
            def wrapper(*args: Any, **kwargs: Any) -> R | None:
                try:
                    return function(*args, **kwargs)
                except PermissionDenied:
                    logger.error(message, exc_info=True)
                    return None

            return wrapper

        message = "Missing permissions to load the configuration"
        config = log_permission_denied(self._installation.load, message=message)(WorkspaceConfig)
        if config is None or config.uber_spn_id is None:
            return

        secret_identifier = f"secrets/{config.inventory_database}/{self._UBER_PRINCIPAL_SECRET_KEY}"
        storage_accounts = self._get_storage_accounts()

        storage_account_ids = ' '.join(str(st.id) for st in storage_accounts)
        message = f"Missing permissions to delete storage permissions for: {storage_account_ids}"
        log_permission_denied(self._azurerm.delete_storage_permission, message=message)(
            config.uber_spn_id, *storage_accounts, safe=True
        )
        message = f"Missing permissions to delete service principal: {config.uber_spn_id}"
        log_permission_denied(self._azurerm.delete_service_principal, message=message)(config.uber_spn_id, safe=True)
        if config.policy_id is not None:
            message = "Missing permissions to revert cluster policy"
            log_permission_denied(self._remove_service_principal_configuration_from_cluster_policy, message=message)(
                config.policy_id, config.uber_spn_id, secret_identifier, storage_accounts
            )
        message = "Missing permissions to revert SQL warehouse config"
        log_permission_denied(
            self._remove_service_principal_configuration_from_workspace_warehouse_config, message=message
        )(config.uber_spn_id, secret_identifier, storage_accounts)
        message = "Missing permissions to delete secret scope"
        log_permission_denied(self._safe_delete_scope, message=message)(config.inventory_database)
        message = "Missing permissions to save the configuration"
        config.uber_spn_id = None
        log_permission_denied(self._installation.save, message=message)(config)

    def _create_access_connector_for_storage_account(
        self, storage_account: StorageAccount, role_name: str = "STORAGE_BLOB_DATA_READER"
    ) -> tuple[AccessConnector, str]:
        access_connector = self._azurerm.create_or_update_access_connector(
            storage_account.id.subscription_id,
            storage_account.id.resource_group,
            f"ac-{storage_account.name}",
            storage_account.location,
            tags={"CreatedBy": "ucx"},
            wait_for_provisioning=True,
        )
        self._apply_storage_permission(access_connector.principal_id, role_name, storage_account)

        container = next(self._azurerm.containers(storage_account.id), None)
        if container is None:
            url = f"abfss://{storage_account.name}.dfs.core.windows.net/"
        else:
            url = f"abfss://{container.container}@{container.storage_account}.dfs.core.windows.net/"

        return access_connector, url

    def create_access_connectors_for_storage_accounts(self) -> list[tuple[AccessConnector, str]]:
        """Create access connectors for storage accounts

        Returns:
            list[AccessConnector, str] : The access connectors with a storage url to which it has access.
        """
        storage_accounts = self._get_storage_accounts()
        if len(storage_accounts) > 200:
            raise RuntimeWarning('Migration will breach UC limits (Storage Credentials > 200).')
        if len(storage_accounts) == 0:
            logger.warning(
                "There are no external table present with azure storage account. "
                "Please check if assessment job is run"
            )
            return []

        tasks = []
        for storage_account in storage_accounts:
            task = partial(
                self._create_access_connector_for_storage_account,
                storage_account=storage_account,
                # Fine-grained access is configured within Databricks through unity
                role_name="STORAGE_BLOB_DATA_CONTRIBUTOR",
            )
            tasks.append(task)

        thread_name = "Creating access connectors for storage accounts"
        results, errors = Threads.gather(thread_name, tasks)
        if len(errors) > 0:
            logger.error(
                "Error creating access connectors. Please review the error message and resolve the issue before trying again. "
                "Removing successfully created access connectors."
            )
            delete_access_connectors = [access_connector for access_connector, _ in results]
            self.delete_access_connectors(*delete_access_connectors)
            raise ManyError(errors)
        return list(results)

    def delete_storage_credential(self, *storage_credentials: Any):
        for storage_credential in storage_credentials:
            self._ws.storage_credentials.delete(storage_credential.name)

    def delete_access_connectors(self, *access_connectors: Any):
        for access_connector in access_connectors:
            self._azurerm.delete_access_connector(str(access_connector.id))

    def _apply_storage_permission(
        self,
        principal_id: str,
        role_name: str,
        *storage_accounts: StorageAccount,
    ):
        for storage in storage_accounts:
            role_guid = str(uuid.uuid4())
            self._azurerm.apply_storage_permission(principal_id, storage, role_name, role_guid)
            logger.debug(f"{role_name} permission applied for spn {principal_id} to storage account {storage.name}")

    def _create_and_get_secret_for_uber_principal(
        self,
        principal_secret: PrincipalSecret,
        scope: str,
    ) -> GetSecretResponse:
        """Create and get a workspace secret for the principal.

        If the secret scope does not, it wil be recreated. If the secret already exists, it will be overwritten.
        """
        logger.info(f"Creating secret scope {scope}.")
        try:
            self._ws.secrets.create_scope(scope)
        except ResourceAlreadyExists:
            logger.warning(f"Secret scope {scope} already exists, using the same")
        self._ws.secrets.put_secret(scope, self._UBER_PRINCIPAL_SECRET_KEY, string_value=principal_secret.secret)
        return self._get_secret(scope, self._UBER_PRINCIPAL_SECRET_KEY)

    @retried(on=[ResourceDoesNotExist], timeout=timedelta(minutes=1))
    def _get_secret(self, scope: str, secret: str) -> GetSecretResponse:
        return self._ws.secrets.get_secret(scope, secret)

    def _safe_delete_scope(self, scope: str) -> None:
        try:
            self._ws.secrets.delete_scope(scope)
        except ResourceDoesNotExist:
            logger.warning(f"Secret scope {scope} does not exist, skipping delete.")
        except PermissionDenied:
            logger.error(f"Missing permissions to delete secret scope: {scope}", exc_info=True)

    def load(self):
        return self._installation.load(list[StoragePermissionMapping], filename=self.FILENAME)

    def _get_storage_accounts(self) -> list[StorageAccount]:
        external_locations = self._locations.snapshot()
        used_storage_accounts = []
        for location in external_locations:
            if location.location.startswith("abfss://"):
                start = location.location.index("@")
                end = location.location.index(".dfs.core.windows.net")
                storage_acct = location.location[start + 1 : end]
                if storage_acct not in used_storage_accounts:
                    used_storage_accounts.append(storage_acct)
        storage_accounts = []
        for storage_account in self._azurerm.storage_accounts():
            if storage_account.name in used_storage_accounts:
                storage_accounts.append(storage_account)
        return storage_accounts



import logging
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Sequence
from typing import ClassVar, Generic, Literal, Protocol, TypeVar

from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk.errors import NotFound

from databricks.labs.ucx.framework.utils import escape_sql_identifier

logger = logging.getLogger(__name__)


class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict]


Result = TypeVar("Result", bound=DataclassInstance)
Dataclass = type[DataclassInstance]
ResultFn = Callable[[], Iterable[Result]]


class CrawlerBase(ABC, Generic[Result]):
    def __init__(self, backend: SqlBackend, catalog: str, schema: str, table: str, klass: type[Result]):
        """
        Initializes a CrawlerBase instance.

        Args:
            backend (SqlBackend): The backend that executes SQL queries:
                Statement Execution API or Databricks Runtime.
            catalog (str): The catalog name for the inventory persistence.
            schema: The schema name for the inventory persistence.
            table: The table name for the inventory persistence.
        """
        self._catalog = self._valid(catalog)
        self._schema = self._valid(schema)
        self._table = self._valid(table)
        self._backend = backend
        self._fetch = backend.fetch
        self._exec = backend.execute
        self._klass = klass

    @property
    def full_name(self) -> str:
        """
        Generates the full name of the table.

        Returns:
            str: The full table name.
        """
        return f"{self._catalog}.{self._schema}.{self._table}"

    def reset(self):
        """
        Delete the content of the inventory table.
        The next call to `snapshot` will re-populate the table.
        """
        self._exec(f"TRUNCATE TABLE {escape_sql_identifier(self.full_name)}")

    @staticmethod
    def _valid(name: str) -> str:
        """
        Validates that the provided name does not contain dots.

        Args:
            name (str): The name to be validated.

        Returns:
            str: The validated name.

        Raises:
            ValueError: If the name contains dots.
        """
        if "." in name:
            msg = f"no dots allowed in `{name}`"
            raise ValueError(msg)
        return name

    @classmethod
    def _try_valid(cls, name: str | None):
        """
        Tries to validate a name. If None, returns None.

        Args:
            name (str): The name to be validated.

        Returns:
            str or None: The validated name or None.
        """
        if name is None:
            return None
        return cls._valid(name)

    def snapshot(self, *, force_refresh: bool = False) -> Iterable[Result]:
        return self._snapshot(self._try_fetch, self._crawl, force_refresh=force_refresh)

    @abstractmethod
    def _try_fetch(self) -> Iterable[Result]:
        """Fetch existing data that has (previously) been crawled by this crawler.

        Returns:
            Iterable[Result]: The data that has already been crawled.
        """

    @abstractmethod
    def _crawl(self) -> Iterable[Result]:
        """Perform the (potentially slow) crawling necessary to capture the current state of the environment.

        Returns:
            Iterable[Result]: Records that capture the results of crawling the environment.
        """

    def _snapshot(self, fetcher: ResultFn, loader: ResultFn, *, force_refresh: bool) -> list[Result]:
        """
        Tries to load dataset of records with `fetcher` function, otherwise automatically creates
        a table with the schema defined in the class of the first row and executes `loader` function
        to populate the dataset.

        Args:
            fetcher: A function to fetch existing data.
            loader: A function to load new data.
            force_refresh: Whether existing data should be ignored and forcibly replaced with data from the loader.

        Exceptions:
        - Any errors raised by the fetcher are passed through, except for NotFound (usually due to
          TABLE_OR_VIEW_NOT_FOUND) which is suppressed. All errors raised by the loader are passed through.

        Returns:
            list[Result]: A list of data records, either fetched or loaded.
        """
        if force_refresh:
            logger.debug(f"[{self.full_name}] ignoring any existing {self._table} inventory; refresh is forced.")
        else:
            logger.debug(f"[{self.full_name}] fetching {self._table} inventory")
            try:
                cached_results = list(fetcher())
                if cached_results:
                    return cached_results
            except NotFound as e:
                logger.debug("Inventory table not found", exc_info=e)
        logger.debug(f"[{self.full_name}] crawling new set of snapshot data for {self._table}")
        loaded_records = list(loader())
        self._update_snapshot(loaded_records, mode="overwrite")
        return loaded_records

    def _update_snapshot(self, items: Sequence[Result], mode: Literal["append", "overwrite"] = "append") -> None:
        logger.debug(f"[{self.full_name}] found {len(items)} new records for {self._table}")
        self._backend.save_table(self.full_name, items, self._klass, mode=mode)



import logging
from collections.abc import Iterator
from dataclasses import replace

from databricks.labs.blueprint.installation import Installation

from databricks.sdk import WorkspaceClient
from databricks.sdk.service.sql import Dashboard, LegacyQuery, UpdateQueryRequestQuery
from databricks.sdk.errors.platform import DatabricksError

from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationIndex
from databricks.labs.ucx.source_code.base import CurrentSessionState
from databricks.labs.ucx.source_code.linters.from_table import FromTableSqlLinter

logger = logging.getLogger(__name__)


class Redash:
    MIGRATED_TAG = "Migrated by UCX"

    def __init__(self, index: TableMigrationIndex, ws: WorkspaceClient, installation: Installation):
        self._index = index
        self._ws = ws
        self._installation = installation

    def migrate_dashboards(self, dashboard_id: str | None = None):
        for dashboard in self._list_dashboards(dashboard_id):
            assert dashboard.id is not None
            if dashboard.tags is not None and self.MIGRATED_TAG in dashboard.tags:
                logger.debug(f"Dashboard {dashboard.name} already migrated by UCX")
                continue
            for query in self.get_queries_from_dashboard(dashboard):
                self._fix_query(query)
            self._ws.dashboards.update(dashboard.id, tags=self._get_migrated_tags(dashboard.tags))

    def revert_dashboards(self, dashboard_id: str | None = None):
        for dashboard in self._list_dashboards(dashboard_id):
            assert dashboard.id is not None
            if dashboard.tags is None or self.MIGRATED_TAG not in dashboard.tags:
                logger.debug(f"Dashboard {dashboard.name} was not migrated by UCX")
                continue
            for query in self.get_queries_from_dashboard(dashboard):
                self._revert_query(query)
            self._ws.dashboards.update(dashboard.id, tags=self._get_original_tags(dashboard.tags))

    def _list_dashboards(self, dashboard_id: str | None) -> list[Dashboard]:
        try:
            if dashboard_id is None:
                return list(self._ws.dashboards.list())
            return [self._ws.dashboards.get(dashboard_id)]
        except DatabricksError as e:
            logger.error(f"Cannot list dashboards: {e}")
            return []

    def _fix_query(self, query: LegacyQuery):
        assert query.id is not None
        assert query.query is not None
        # query already migrated
        if query.tags is not None and self.MIGRATED_TAG in query.tags:
            return
        # backup the query
        self._installation.save(query, filename=f'backup/queries/{query.id}.json')
        from_table = FromTableSqlLinter(self._index, self._get_session_state(query))
        new_query = UpdateQueryRequestQuery(
            query_text=from_table.apply(query.query),
            tags=self._get_migrated_tags(query.tags),
        )
        try:
            self._ws.queries.update(
                query.id,
                update_mask="query_text,tags",
                query=new_query,
            )
        except DatabricksError:
            logger.warning(f"Cannot upgrade {query.name}")
            return

    @staticmethod
    def _get_session_state(query: LegacyQuery) -> CurrentSessionState:
        session_state = CurrentSessionState()
        if query.options is None:
            return session_state
        if query.options.catalog:
            session_state = replace(session_state, catalog=query.options.catalog)
        if query.options.schema:
            session_state = replace(session_state, schema=query.options.schema)
        return session_state

    def _revert_query(self, query: LegacyQuery):
        assert query.id is not None
        assert query.query is not None
        if query.tags is None:
            return
        # find the backup query
        is_migrated = False
        for tag in query.tags:
            if tag == self.MIGRATED_TAG:
                is_migrated = True

        if not is_migrated:
            logger.debug(f"Query {query.name} was not migrated by UCX")
            return

        backup_query = self._installation.load(LegacyQuery, filename=f'backup/queries/{query.id}.json')
        update_query = UpdateQueryRequestQuery(
            query_text=backup_query.query, tags=self._get_original_tags(backup_query.tags)
        )
        try:
            self._ws.queries.update(query.id, update_mask="query_text,tags", query=update_query)
        except DatabricksError:
            logger.warning(f"Cannot restore {query.name} from backup")
            return

    def _get_migrated_tags(self, tags: list[str] | None) -> list[str]:
        out = [self.MIGRATED_TAG]
        if tags:
            out.extend(tags)
        return out

    def _get_original_tags(self, tags: list[str] | None) -> list[str] | None:
        if tags is None:
            return None
        return [tag for tag in tags if tag != self.MIGRATED_TAG]

    @staticmethod
    def get_queries_from_dashboard(dashboard: Dashboard) -> Iterator[LegacyQuery]:
        if dashboard.widgets is None:
            return
        for widget in dashboard.widgets:
            if widget is None:
                continue
            if widget.visualization is None:
                continue
            if widget.visualization.query is None:
                continue
            yield widget.visualization.query



import dataclasses
import logging
from collections.abc import Iterable
from dataclasses import asdict, dataclass
from datetime import datetime, timezone

from databricks.sdk import WorkspaceClient
from databricks.sdk.service.sql import Dashboard, LegacyQuery
from databricks.sdk.service.workspace import Language

from databricks.labs.lsql.backends import SqlBackend

from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationIndex
from databricks.labs.ucx.source_code.base import CurrentSessionState
from databricks.labs.ucx.source_code.directfs_access import DirectFsAccessCrawler, DirectFsAccess, LineageAtom
from databricks.labs.ucx.source_code.linters.context import LinterContext
from databricks.labs.ucx.source_code.linters.directfs import DirectFsAccessSqlLinter
from databricks.labs.ucx.source_code.redash import Redash

logger = logging.getLogger(__name__)


@dataclass
class QueryProblem:
    dashboard_id: str
    dashboard_parent: str
    dashboard_name: str
    query_id: str
    query_parent: str
    query_name: str
    code: str
    message: str


class QueryLinter:

    def __init__(
        self,
        ws: WorkspaceClient,
        migration_index: TableMigrationIndex,
        directfs_crawler: DirectFsAccessCrawler,
    ):
        self._ws = ws
        self._migration_index = migration_index
        self._directfs_crawler = directfs_crawler

    def refresh_report(self, sql_backend: SqlBackend, inventory_database: str):
        assessment_start = datetime.now(timezone.utc)
        linted_queries: set[str] = set()
        all_dashboards = list(self._ws.dashboards.list())
        logger.info(f"Running {len(all_dashboards)} linting tasks...")
        all_problems: list[QueryProblem] = []
        all_dfsas: list[DirectFsAccess] = []
        # first lint and collect queries from dashboards
        for dashboard in all_dashboards:
            if not dashboard.id:
                continue
            dashboard = self._ws.dashboards.get(dashboard_id=dashboard.id)
            problems, dfsas = self._lint_and_collect_from_dashboard(dashboard, linted_queries)
            all_problems.extend(problems)
            all_dfsas.extend(dfsas)
        for query in self._ws.queries_legacy.list():
            if query.id is None:
                continue
            if query.id in linted_queries:
                continue
            linted_queries.add(query.id)
            problems = self.lint_query(query)
            all_problems.extend(problems)
            dfsas = self.collect_dfsas_from_query(query)
            all_dfsas.extend(dfsas)
        # dump problems
        logger.info(f"Saving {len(all_problems)} linting problems...")
        sql_backend.save_table(
            f'{escape_sql_identifier(inventory_database)}.query_problems',
            all_problems,
            QueryProblem,
            mode='overwrite',
        )
        # dump dfsas
        assessment_end = datetime.now(timezone.utc)
        all_dfsas = [
            dataclasses.replace(
                dfsa, assessment_start_timestamp=assessment_start, assessment_end_timestamp=assessment_end
            )
            for dfsa in all_dfsas
        ]
        self._directfs_crawler.dump_all(all_dfsas)

    def _lint_and_collect_from_dashboard(
        self, dashboard: Dashboard, linted_queries: set[str]
    ) -> tuple[Iterable[QueryProblem], Iterable[DirectFsAccess]]:
        dashboard_queries = Redash.get_queries_from_dashboard(dashboard)
        query_problems: list[QueryProblem] = []
        query_dfsas: list[DirectFsAccess] = []
        dashboard_id = dashboard.id or "<no-id>"
        dashboard_parent = dashboard.parent or "<orphan>"
        dashboard_name = dashboard.name or "<anonymous>"
        for query in dashboard_queries:
            if query.id is None:
                continue
            if query.id in linted_queries:
                continue
            linted_queries.add(query.id)
            problems = self.lint_query(query)
            for problem in problems:
                query_problems.append(
                    dataclasses.replace(
                        problem,
                        dashboard_id=dashboard_id,
                        dashboard_parent=dashboard_parent,
                        dashboard_name=dashboard_name,
                    )
                )
            dfsas = self.collect_dfsas_from_query(query)
            for dfsa in dfsas:
                atom = LineageAtom(
                    object_type="DASHBOARD",
                    object_id=dashboard_id,
                    other={"parent": dashboard_parent, "name": dashboard_name},
                )
                source_lineage = [atom] + dfsa.source_lineage
                query_dfsas.append(dataclasses.replace(dfsa, source_lineage=source_lineage))
        return query_problems, query_dfsas

    def lint_query(self, query: LegacyQuery) -> Iterable[QueryProblem]:
        if not query.query:
            return
        ctx = LinterContext(self._migration_index, CurrentSessionState())
        linter = ctx.linter(Language.SQL)
        query_id = query.id or "<no-id>"
        query_parent = query.parent or "<orphan>"
        query_name = query.name or "<anonymous>"
        for advice in linter.lint(query.query):
            yield QueryProblem(
                dashboard_id="",
                dashboard_parent="",
                dashboard_name="",
                query_id=query_id,
                query_parent=query_parent,
                query_name=query_name,
                code=advice.code,
                message=advice.message,
            )

    @classmethod
    def collect_dfsas_from_query(cls, query: LegacyQuery) -> Iterable[DirectFsAccess]:
        if query.query is None:
            return
        linter = DirectFsAccessSqlLinter()
        source_id = query.id or "no id"
        source_name = query.name or "<anonymous>"
        source_timestamp = cls._read_timestamp(query.updated_at)
        source_lineage = [LineageAtom(object_type="QUERY", object_id=source_id, other={"name": source_name})]
        for dfsa in linter.collect_dfsas(query.query):
            yield DirectFsAccess(**asdict(dfsa)).replace_source(
                source_id=source_id, source_timestamp=source_timestamp, source_lineage=source_lineage
            )

    @classmethod
    def _read_timestamp(cls, timestamp: str | None) -> datetime:
        if timestamp is not None:
            methods = [
                datetime.fromisoformat,
                lambda s: datetime.fromisoformat(s[:-1]),  # ipython breaks on final 'Z'
            ]
            for method in methods:
                try:
                    return method(timestamp)
                except ValueError:
                    pass
        return datetime.now()



import logging
from dataclasses import dataclass

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.tui import Prompts
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors.platform import InvalidParameterValue
from databricks.sdk.service.catalog import (
    AwsIamRoleRequest,
    Privilege,
    StorageCredentialInfo,
    ValidationResultResult,
)

from databricks.labs.ucx.assessment.aws import AWSRoleAction, AWSUCRoleCandidate, AWSCredentialCandidate
from databricks.labs.ucx.aws.access import AWSResourcePermissions

logger = logging.getLogger(__name__)


@dataclass
class CredentialValidationResult:
    name: str
    role_arn: str
    validated_on: str
    read_only: bool
    failures: list[str] | None = None


class CredentialManager:
    def __init__(self, ws: WorkspaceClient):
        self._ws = ws

    def list(self, include_names: set[str] | None = None) -> set[str]:
        # list existed storage credentials that is using iam roles, capturing the arns
        iam_roles = set()

        storage_credentials = self._ws.storage_credentials.list(max_results=0)

        for storage_credential in storage_credentials:

            # only add storage credentials with iam roles
            if not storage_credential.aws_iam_role:
                continue

            # if include_names is not None, only add the storage credentials that are in the include_names set
            if include_names is not None and storage_credential.name not in include_names:
                continue

            iam_roles.add(storage_credential.aws_iam_role.role_arn)

        logger.info(f"Found {len(iam_roles)} distinct IAM roles already used in UC storage credentials")
        return iam_roles

    def create(self, name: str, role_arn: str, read_only: bool) -> StorageCredentialInfo:
        return self._ws.storage_credentials.create(
            name,
            aws_iam_role=AwsIamRoleRequest(role_arn),
            comment=f"Created by UCX during migration to UC using AWS IAM Role: {name}",
            read_only=read_only,
        )

    def validate(self, role_action: AWSRoleAction) -> CredentialValidationResult:
        try:
            validation = self._ws.storage_credentials.validate(
                storage_credential_name=role_action.role_name,
                url=role_action.resource_path,
                read_only=role_action.privilege == Privilege.READ_FILES.value,
            )
        except InvalidParameterValue:
            logger.warning(
                "There is an existing external location overlaps with the prefix that is mapped to "
                "the IAM Role and used for validating the migrated storage credential. "
                "Skip the validation"
            )
            return CredentialValidationResult(
                role_action.role_name,
                role_action.role_arn,
                role_action.resource_path,
                role_action.privilege == Privilege.READ_FILES.value,
                [
                    "The validation is skipped because an existing external location overlaps "
                    "with the location used for validation."
                ],
            )

        if not validation.results:
            return CredentialValidationResult(
                role_action.role_name,
                role_action.role_arn,
                role_action.resource_path,
                role_action.privilege == Privilege.READ_FILES.value,
                ["Validation returned no results."],
            )

        failures = []
        for result in validation.results:
            if result.operation is None:
                continue
            if result.result == ValidationResultResult.FAIL:
                failures.append(f"{result.operation.value} validation failed with message: {result.message}")
        return CredentialValidationResult(
            role_action.role_name,
            role_action.role_arn,
            role_action.resource_path,
            role_action.privilege == Privilege.READ_FILES.value,
            None if not failures else failures,
        )


class IamRoleMigration:

    def __init__(
        self,
        installation: Installation,
        resource_permissions: AWSResourcePermissions,
        storage_credential_manager: CredentialManager,
    ):
        self._output_file = "aws_iam_role_migration_result.csv"
        self._installation = installation
        self._resource_permissions = resource_permissions
        self._storage_credential_manager = storage_credential_manager

    @staticmethod
    def _print_action_plan(iam_list: list[AWSCredentialCandidate]):
        # print action plan to console for customer to review.
        for iam in iam_list:
            logger.info(f"Credential {iam.role_name} --> {iam.role_arn}: {iam.privilege}")

    def _generate_migration_list(self, include_names: set[str] | None = None) -> list[AWSCredentialCandidate]:
        """
        Create the list of IAM roles that need to be migrated, output an action plan as a csv file for users to confirm.
        It returns a list of ARNs
        """
        # load IAM role list
        iam_list = self._resource_permissions.get_roles_to_migrate()
        # list existing storage credentials
        sc_set = self._storage_credential_manager.list(include_names)
        # check if the iam is already used in UC storage credential
        filtered_iam_list = [iam for iam in iam_list if iam.role_arn not in sc_set]

        # output the action plan for customer to confirm
        self._print_action_plan(filtered_iam_list)

        return filtered_iam_list

    def save(self, migration_results: list[CredentialValidationResult]) -> str:
        return self._installation.save(migration_results, filename=self._output_file)

    def run(self, prompts: Prompts, include_names: set[str] | None = None) -> list[CredentialValidationResult]:

        iam_list = self._generate_migration_list(include_names)
        if len(iam_list) == 0:
            logger.info("No IAM Role to migrate")
            return []

        plan_confirmed = prompts.confirm(
            "Above IAM roles will be migrated to UC storage credentials, please review and confirm."
        )
        if plan_confirmed is not True:
            return []

        execution_result = []
        for iam in iam_list:
            storage_credential = self._storage_credential_manager.create(
                name=iam.role_name,
                role_arn=iam.role_arn,
                read_only=iam.privilege == Privilege.READ_FILES.value,
            )
            if storage_credential.aws_iam_role is None:
                logger.error(f"Failed to create storage credential for IAM role: {iam.role_arn}")
                continue
            self._resource_permissions.update_uc_role(
                iam.role_name, iam.role_arn, storage_credential.aws_iam_role.external_id
            )
            for path in iam.paths:
                role_action = AWSRoleAction(iam.role_arn, "s3", path, iam.privilege)
                execution_result.append(self._storage_credential_manager.validate(role_action))

        if execution_result:
            results_file = self.save(execution_result)
            logger.info(
                f"Completed migration from IAM Role to UC Storage credentials. "
                f"Please check {results_file} for validation results"
            )
        else:
            logger.info("No IAM Role migrated to UC Storage credentials")
        return execution_result


class IamRoleCreation:

    def __init__(
        self,
        installation: Installation,
        ws: WorkspaceClient,
        resource_permissions: AWSResourcePermissions,
    ):
        self._output_file = "aws_iam_role_creation_result.csv"
        self._installation = installation
        self._ws = ws
        self._resource_permissions = resource_permissions

    @staticmethod
    def _print_action_plan(iam_list: list[AWSUCRoleCandidate]):
        # print action plan to console for customer to review.
        for iam in iam_list:
            logger.info(f"Role:{iam.role_name} Policy:{iam.policy_name} Paths:{iam.resource_paths}")

    def save(self, migration_results: list[CredentialValidationResult]) -> str:
        return self._installation.save(migration_results, filename=self._output_file)

    def run(self, prompts: Prompts, *, single_role=False, role_name="UC_ROLE", policy_name="UC_POLICY"):

        iam_list = self._resource_permissions.list_uc_roles(
            single_role=single_role,
            role_name=role_name,
            policy_name=policy_name,
        )
        if not iam_list:
            logger.info("No IAM Role created")
            return
        self._print_action_plan(iam_list)
        plan_confirmed = prompts.confirm(
            "Above UC Compatible IAM roles will be created and granted access to the corresponding paths, "
            "please review and confirm."
        )
        if plan_confirmed is not True:
            return

        self._resource_permissions.create_uc_roles(iam_list)
        self._resource_permissions.save_uc_compatible_roles()

    def delete_uc_roles(self, prompts: Prompts) -> None:
        uc_roles = self._resource_permissions.load_uc_compatible_roles()
        if len(uc_roles) == 0:
            self._resource_permissions.save_uc_compatible_roles()
            uc_roles = self._resource_permissions.load_uc_compatible_roles()
        storage_credentials = self._ws.storage_credentials.list()

        uc_role_mapping = {role.role_arn: role.role_name for role in uc_roles}
        selected_roles = prompts.multiple_choice_from_dict("Select the list of roles created by UCX", uc_role_mapping)
        if len(selected_roles) == 0:
            logger.info("No roles selected...")
            return
        matching_credentials = []
        for storage_credential in storage_credentials:
            if (
                storage_credential.aws_iam_role is not None
                and uc_role_mapping.get(storage_credential.aws_iam_role.role_arn) in selected_roles
            ):
                matching_credentials.append(storage_credential)

        for credential in matching_credentials:
            if credential.aws_iam_role is not None:
                logger.info(f"Storage credential: {credential.name} IAM Role: {credential.aws_iam_role.role_arn}")
        if len(matching_credentials) == 0:
            logger.info("No storage credential using the selected UC roles, proceeding to delete.")
        if len(matching_credentials) == 0 or prompts.confirm(
            "The above storage credential will be impacted on deleting the selected IAM roles,"
            " Are you sure you want to confirm"
        ):

            logger.info("Deleting UCX created roles...")
            for role_name in selected_roles:
                logger.info(f"Deleting role {role_name}.")
                self._resource_permissions.delete_uc_role(role_name)



from __future__ import annotations

import logging
import sys
from pathlib import Path


logger = logging.getLogger(__name__)


class PathLookup:
    """
    Mimic Python's importlib Lookup.

    Sources:
        See Lookup in importlib.metadata.__init__.py
    """

    @classmethod
    def from_pathlike_string(cls, cwd: Path, syspath: str):
        paths = syspath.split(':')
        return PathLookup(cwd, [Path(path) for path in paths])

    @classmethod
    def from_sys_path(cls, cwd: Path):
        return PathLookup(cwd, [Path(path) for path in sys.path])

    def __init__(self, cwd: Path, sys_paths: list[Path]):
        self._cwd = cwd
        self._sys_paths = sys_paths

    def change_directory(self, new_working_directory: Path) -> PathLookup:
        return PathLookup(new_working_directory, self._sys_paths)

    def resolve(self, path: Path) -> Path | None:
        try:
            if path.is_absolute() and path.exists():
                # eliminate ".." components
                return path.resolve()
        except PermissionError:
            logger.warning(f"Permission denied to access {path}")
            return None
        for library_root in self.library_roots:
            try:
                resolved_path = self._resolve_in_library_root(library_root, path)
            except PermissionError:
                logger.warning(f"Permission denied to access files or directories in {library_root}")
                continue
            if resolved_path is not None:
                return resolved_path
        return None

    def _resolve_in_library_root(self, library_root: Path, path: Path) -> Path | None:
        if not library_root.is_dir():
            return None
        absolute_path = library_root / path
        if absolute_path.exists():
            return self._standardize_path(absolute_path)
        return self._resolve_egg_in_library_root(library_root, path)

    def _resolve_egg_in_library_root(self, library: Path, path: Path) -> Path | None:
        for child in library.iterdir():
            if not self._is_egg_folder(child):
                continue
            absolute_path = child / path
            if absolute_path.exists():
                return self._standardize_path(absolute_path)
        return None

    @staticmethod
    def _standardize_path(path: Path):
        resolved = path.resolve()  # eliminate ".." components
        # on MacOS "/var/..." resolves to "/private/var/.." which breaks path equality
        index = 1 if path.parts[0] == path.anchor else 0
        if path.parts[index] == resolved.parts[index]:
            return resolved
        posix = resolved.as_posix()
        posix = posix[posix.index(path.as_posix()) :]
        return Path(posix)

    @staticmethod
    def _is_egg_folder(path: Path) -> bool:
        """Egg folders end with `.egg` and have a 'EGG-INFO' file."""
        return (
            path.is_dir()
            and path.suffix == ".egg"
            and any(subfolder.name.lower() == "egg-info" for subfolder in path.iterdir())
        )

    def has_path(self, path: Path):
        return next(p for p in self._sys_paths if path == p) is not None

    def prepend_path(self, path: Path):
        self._sys_paths.insert(0, path)

    def insert_path(self, index: int, path: Path):
        self._sys_paths.insert(index, path)

    def append_path(self, path: Path):
        if path in self._sys_paths:
            return
        self._sys_paths.append(path)

    def remove_path(self, index: int):
        del self._sys_paths[index]

    @property
    def library_roots(self) -> list[Path]:
        # we may return a combination of DBFSPath, WorkspacePath and PosixPath here
        library_roots = []
        for library_root in [self._cwd] + self._sys_paths:
            try:
                is_existing_directory = library_root.exists() and library_root.is_dir()
            except PermissionError:
                continue
            if is_existing_directory:
                library_roots.append(library_root)
        return library_roots

    @property
    def cwd(self):
        return self._cwd

    def __repr__(self):
        return f"PathLookup(cwd={self._cwd}, sys_paths={self._sys_paths})"



import base64
import json
import logging
from collections.abc import Iterable
from dataclasses import dataclass

from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import ResourceDoesNotExist

from databricks.labs.ucx.assessment.crawlers import (
    AZURE_SP_CONF_FAILURE_MSG,
    azure_sp_conf_in_init_scripts,
)
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.framework.crawlers import CrawlerBase

logger = logging.getLogger(__name__)


@dataclass
class GlobalInitScriptInfo:
    script_id: str
    success: int
    failures: str
    script_name: str | None = None
    created_by: str | None = None
    enabled: bool | None = None


class CheckInitScriptMixin:
    _ws: WorkspaceClient

    def check_init_script(self, init_script_data: str | None, source: str) -> list[str]:
        failures: list[str] = []
        if not init_script_data:
            return failures
        if azure_sp_conf_in_init_scripts(init_script_data):
            failures.append(f"{AZURE_SP_CONF_FAILURE_MSG} {source}.")
        return failures


class GlobalInitScriptCrawler(CrawlerBase[GlobalInitScriptInfo], CheckInitScriptMixin):
    def __init__(self, ws: WorkspaceClient, sbe: SqlBackend, schema):
        super().__init__(sbe, "hive_metastore", schema, "global_init_scripts", GlobalInitScriptInfo)
        self._ws = ws

    def _crawl(self) -> Iterable[GlobalInitScriptInfo]:
        all_global_init_scripts = list(self._ws.global_init_scripts.list())
        return list(self._assess_global_init_scripts(all_global_init_scripts))

    def _assess_global_init_scripts(self, all_global_init_scripts):
        for gis in all_global_init_scripts:
            if not gis.created_by:
                logger.warning(
                    f"Script {gis.name} have Unknown creator, it means that the original creator has been deleted"
                    f" and should be re-created"
                )
            global_init_script_info = GlobalInitScriptInfo(
                script_id=gis.script_id,
                script_name=gis.name,
                created_by=gis.created_by,
                enabled=gis.enabled,
                success=1,
                failures="[]",
            )
            failures = []
            try:
                script = self._ws.global_init_scripts.get(gis.script_id)
            except ResourceDoesNotExist:
                logger.warning(f"removed on the backend {gis.script_id}")
                continue
            global_init_script = base64.b64decode(script.script).decode("utf-8")
            if not global_init_script:
                continue
            failures.extend(self.check_init_script(global_init_script, "global init script"))
            global_init_script_info.failures = json.dumps(failures)

            if len(failures) > 0:
                global_init_script_info.success = 0
            yield global_init_script_info

    def _try_fetch(self) -> Iterable[GlobalInitScriptInfo]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield GlobalInitScriptInfo(*row)



from __future__ import annotations

import codecs
import dataclasses
import locale
import logging
from abc import abstractmethod, ABC
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path

from astroid import AstroidSyntaxError, NodeNG  # type: ignore
from sqlglot import Expression, parse as parse_sql, ParseError as SqlParseError

from databricks.sdk.service import compute
from databricks.sdk.service.workspace import Language

from databricks.labs.blueprint.paths import WorkspacePath

from databricks.labs.ucx.source_code.python.python_ast import Tree

# Code mapping between LSP, PyLint, and our own diagnostics:
# | LSP                       | PyLint     | Our            |
# |---------------------------|------------|----------------|
# | Severity.ERROR            | Error      | Failure()      |
# | Severity.WARN             | Warning    | Advisory()     |
# | DiagnosticTag.DEPRECATED  | Warning    | Deprecation()  |
# | Severity.INFO             | Info       | Advice()       |
# | Severity.HINT             | Convention | Convention()   |
# | DiagnosticTag.UNNECESSARY | Refactor   | Convention()   |

logger = logging.getLogger(__name__)


@dataclass
class Advice:
    code: str
    message: str
    # Lines and columns are both 0-based: the first line is line 0.
    start_line: int
    start_col: int
    end_line: int
    end_col: int

    def as_advisory(self) -> 'Advisory':
        return Advisory(**self.__dict__)

    def as_failure(self) -> 'Failure':
        return Failure(**self.__dict__)

    def as_deprecation(self) -> 'Deprecation':
        return Deprecation(**self.__dict__)

    def as_convention(self) -> 'Convention':
        return Convention(**self.__dict__)

    def for_path(self, path: Path) -> LocatedAdvice:
        return LocatedAdvice(self, path)

    @classmethod
    def from_node(cls, *, code: str, message: str, node: NodeNG) -> Advice:
        # Astroid lines are 1-based.
        return cls(
            code=code,
            message=message,
            start_line=(node.lineno or 1) - 1,
            start_col=node.col_offset or 0,
            end_line=(node.end_lineno or 1) - 1,
            end_col=node.end_col_offset,
        )

    def replace_from_node(self, node: NodeNG) -> Advice:
        # Astroid lines are 1-based.
        return dataclasses.replace(
            self,
            start_line=(node.lineno or 1) - 1,
            start_col=node.col_offset,
            end_line=(node.end_lineno or 1) - 1,
            end_col=node.end_col_offset,
        )


@dataclass
class LocatedAdvice:
    advice: Advice
    path: Path

    @property
    def is_unknown(self):
        return self.path == Path('UNKNOWN')

    def message_relative_to(self, base: Path, *, default: Path | None = None) -> str:
        advice = self.advice
        path = self.path
        if self.is_unknown:
            logger.debug(f'THIS IS A BUG! {advice.code}:{advice.message} has unknown path')
        if default is not None:
            path = default
        path = path.relative_to(base)
        return f"./{path.as_posix()}:{advice.start_line}:{advice.start_col}: [{advice.code}] {advice.message}"


class Advisory(Advice):
    """A warning that does not prevent the code from running."""


class Failure(Advice):
    """An error that prevents the code from running."""


class Deprecation(Advice):
    """An advisory that suggests to replace the code with a newer version."""


class Convention(Advice):
    """A suggestion for a better way to write the code."""


class Linter:
    @abstractmethod
    def lint(self, code: str) -> Iterable[Advice]: ...


class SqlLinter(Linter):

    def lint(self, code: str) -> Iterable[Advice]:
        try:
            expressions = parse_sql(code, read='databricks')
            for expression in expressions:
                if not expression:
                    continue
                yield from self.lint_expression(expression)
        except SqlParseError as e:
            logger.debug(f"Failed to parse SQL: {code}", exc_info=e)
            yield self.sql_parse_failure(code)

    @staticmethod
    def sql_parse_failure(code: str):
        return Failure(
            code='sql-parse-error',
            message=f"SQL expression is not supported yet: {code}",
            # SQLGlot does not propagate tokens yet. See https://github.com/tobymao/sqlglot/issues/3159
            start_line=0,
            start_col=0,
            end_line=0,
            end_col=1024,
        )

    @abstractmethod
    def lint_expression(self, expression: Expression) -> Iterable[Advice]: ...


class PythonLinter(Linter):

    def lint(self, code: str) -> Iterable[Advice]:
        tree = Tree.normalize_and_parse(code)
        yield from self.lint_tree(tree)

    @abstractmethod
    def lint_tree(self, tree: Tree) -> Iterable[Advice]: ...


class Fixer(ABC):

    @property
    @abstractmethod
    def name(self) -> str: ...

    @abstractmethod
    def apply(self, code: str) -> str: ...


# The default schema to use when the schema is not specified in a table reference
# See: https://spark.apache.org/docs/3.0.0-preview/sql-ref-syntax-qry-select-usedb.html
DEFAULT_CATALOG = 'hive_metastore'
DEFAULT_SCHEMA = 'default'


@dataclass
class CurrentSessionState:
    """
    A data class that represents the current state of a session.

    This class can be used to track various aspects of a session, such as the current schema.

    Attributes:
        catalog (str): The current schema of the session. If not provided, it defaults to 'DEFAULT_CATALOG'.
        schema (str): The current schema of the session. If not provided, it defaults to 'DEFAULT_SCHEMA'.
    """

    schema: str = DEFAULT_SCHEMA
    catalog: str = DEFAULT_CATALOG
    spark_conf: dict[str, str] | None = None
    named_parameters: dict[str, str] | None = None
    data_security_mode: compute.DataSecurityMode | None = None
    is_serverless: bool = False
    dbr_version: tuple[int, int] | None = None

    @classmethod
    def from_json(cls, json: dict) -> CurrentSessionState:
        return cls(
            schema=json.get('schema', DEFAULT_SCHEMA),
            catalog=json.get('catalog', DEFAULT_CATALOG),
            spark_conf=json.get('spark_conf', None),
            named_parameters=json.get('named_parameters', None),
            data_security_mode=cls.parse_security_mode(json.get('data_security_mode', None)),
            is_serverless=json.get('is_serverless', False),
            dbr_version=tuple(json['dbr_version']) if 'dbr_version' in json else None,
        )

    @staticmethod
    def parse_security_mode(mode_str: str | None) -> compute.DataSecurityMode | None:
        try:
            return compute.DataSecurityMode(mode_str) if mode_str else None
        except ValueError:
            logger.warning(f'Unknown data_security_mode {mode_str}')
            return None


class SqlSequentialLinter(SqlLinter):

    def __init__(self, linters: list[SqlLinter]):
        self._linters = linters

    def lint_expression(self, expression: Expression) -> Iterable[Advice]:
        for linter in self._linters:
            yield from linter.lint_expression(expression)


class PythonSequentialLinter(Linter):

    def __init__(self, linters: list[PythonLinter]):
        self._linters = linters
        self._tree: Tree | None = None

    def lint(self, code: str) -> Iterable[Advice]:
        try:
            tree = self._parse_and_append(code)
            yield from self.lint_tree(tree)
        except AstroidSyntaxError as e:
            yield Failure('syntax-error', str(e), 0, 0, 0, 0)

    def lint_tree(self, tree: Tree) -> Iterable[Advice]:
        for linter in self._linters:
            yield from linter.lint_tree(tree)

    def _parse_and_append(self, code: str) -> Tree:
        tree = Tree.normalize_and_parse(code)
        self.append_tree(tree)
        return tree

    def append_tree(self, tree: Tree):
        self._make_tree().append_tree(tree)

    def append_nodes(self, nodes: list[NodeNG]):
        self._make_tree().append_nodes(nodes)

    def append_globals(self, globs: dict):
        self._make_tree().append_globals(globs)

    def process_child_cell(self, code: str):
        try:
            this_tree = self._make_tree()
            tree = Tree.normalize_and_parse(code)
            this_tree.append_tree(tree)
        except AstroidSyntaxError as e:
            # error already reported when linting enclosing notebook
            logger.warning(f"Failed to parse Python cell: {code}", exc_info=e)

    def _make_tree(self) -> Tree:
        if self._tree is None:
            self._tree = Tree.new_module()
        return self._tree


SUPPORTED_EXTENSION_LANGUAGES = {
    '.py': Language.PYTHON,
    '.sql': Language.SQL,
}


def file_language(path: Path) -> Language | None:
    return SUPPORTED_EXTENSION_LANGUAGES.get(path.suffix.lower())


def guess_encoding(path: Path):
    # some files encode a unicode BOM (byte-order-mark), so let's use that if available
    with path.open('rb') as _file:
        raw = _file.read(4)
        if raw.startswith(codecs.BOM_UTF32_LE) or raw.startswith(codecs.BOM_UTF32_BE):
            return 'utf-32'
        if raw.startswith(codecs.BOM_UTF16_LE) or raw.startswith(codecs.BOM_UTF16_BE):
            return 'utf-16'
        if raw.startswith(codecs.BOM_UTF8):
            return 'utf-8-sig'
        # no BOM, let's use default encoding
        return locale.getpreferredencoding(False)


# duplicated from CellLanguage to prevent cyclic import
LANGUAGE_COMMENT_PREFIXES = {Language.PYTHON: '#', Language.SCALA: '//', Language.SQL: '--'}
NOTEBOOK_HEADER = "Databricks notebook source"


def is_a_notebook(path: Path, content: str | None = None) -> bool:
    if isinstance(path, WorkspacePath):
        return path.is_notebook()
    if not path.is_file():
        return False
    language = file_language(path)
    if not language:
        return False
    magic_header = f"{LANGUAGE_COMMENT_PREFIXES.get(language)} {NOTEBOOK_HEADER}"
    if content is not None:
        return content.startswith(magic_header)
    try:
        with path.open('rt', encoding=guess_encoding(path)) as f:
            file_header = f.read(len(magic_header))
    except (FileNotFoundError, UnicodeDecodeError, PermissionError):
        logger.warning(f"Could not read file {path}")
        return False
    return file_header == magic_header
